#
#  test_variables
#

import dataclasses

import pandas as pd
import pytest

from owid.catalog.meta import VariableMeta, VariablePresentationMeta
from owid.catalog.variables import (
    License,
    Variable,
    combine_variables_metadata,
    get_unique_licenses_from_variables,
    get_unique_origins_from_variables,
    get_unique_sources_from_variables,
)


def test_create_empty_variable() -> None:
    v = Variable(name="dog")
    assert v is not None


def test_create_unnamed_variable_fails() -> None:
    v = Variable()
    assert v.name is None

    # cannot access a metadata attribute without a name
    with pytest.raises(ValueError):
        v.title

    # cannot set a metadata attribute without a name
    with pytest.raises(ValueError):
        v.description = "Hello"


def test_operations_on_unnamed_variables_succeed() -> None:
    # we currently rely on this for interim logic in the ETL
    v = Variable([1, 2, 3])
    v2 = v + 1  # noqa


def test_create_non_empty_variable() -> None:
    v = Variable([1, 2, 3], name="dog")
    assert list(v) == [1, 2, 3]


def test_metadata_survives_slicing() -> None:
    v = Variable([1, 2, 3], name="dog")
    v.description = "Something amazing"

    assert isinstance(v.iloc[:2], Variable)
    # assert v.iloc[:2].description == v.description


def test_metadata_accessed_in_bulk() -> None:
    v = Variable([1, 2, 3], name="dog")
    assert v.metadata == VariableMeta()


def test_variable_can_be_type_cast() -> None:
    v = Variable([1, 2, 3], name="dog", dtype="object")
    v.metadata.description = "Blah blah..."
    v2 = v.astype("int")
    assert v2.name == v.name
    assert v2.metadata == v.metadata
    assert (v == v2).all()


def _assert_untouched_data_and_metadata_did_not_change(tb1, tb1_expected):
    # Check that all columns that were in the old table have not been affected.
    for column in tb1_expected.columns:
        assert (tb1[column] == tb1_expected[column]).all()
        assert tb1._fields[column] == tb1_expected._fields[column]


def test_create_new_variable_as_sum_of_other_two(table_1, sources, origins, licenses) -> None:
    tb1 = table_1.copy()
    tb1["c"] = tb1["a"] + tb1["b"]
    _assert_untouched_data_and_metadata_did_not_change(tb1=tb1, tb1_expected=table_1)
    assert (tb1["c"] == pd.Series([5, 7, 9])).all()
    # Since "a" and "b" have different title and description, "c" should have no title or description.
    assert tb1["c"].metadata.title is None
    assert tb1["c"].metadata.description is None
    assert tb1["c"].metadata.description_key == [
        "Key description point 1 of Variable 1",
        "Common key description point",
        "Key description point 1 of Variable 2",
        "Key description point 2 of Variable 2",
    ]
    assert tb1["c"].metadata.sources == [sources[2], sources[1], sources[3]]
    assert tb1["c"].metadata.origins == [origins[2], origins[1], origins[3]]
    assert tb1["c"].metadata.licenses == [licenses[1], licenses[2], licenses[3]]
    # Processing level should be the highest of the two variables.
    assert tb1["c"].metadata.processing_level == "major"
    # Both "a" and "b" have different values in presentation, the combination should have no presentation.
    assert tb1["c"].metadata.presentation is None
    # Since "a" and "b" have identical display, the combination should have the same display.
    assert tb1["c"].metadata.display == tb1["a"].metadata.display


def test_create_new_variable_as_sum_of_another_variable_plus_a_scalar(table_1) -> None:
    tb1 = table_1.copy()
    tb1["d"] = tb1["a"] + 1
    _assert_untouched_data_and_metadata_did_not_change(tb1=tb1, tb1_expected=table_1)
    assert (tb1["d"] == pd.Series([2, 3, 4])).all()
    assert tb1["d"].metadata.title == table_1["a"].metadata.title
    assert tb1["d"].metadata.description == table_1["a"].metadata.description
    assert tb1["d"].metadata.description_key == [
        "Key description point 1 of Variable 1",
        "Common key description point",
    ]
    assert tb1["d"].metadata.sources == table_1["a"].metadata.sources
    assert tb1["d"].metadata.origins == table_1["a"].metadata.origins
    assert tb1["d"].metadata.licenses == table_1["a"].metadata.licenses
    assert tb1["d"].metadata.processing_level == "minor"
    assert tb1["d"].metadata.presentation == tb1["a"].metadata.presentation
    assert tb1["d"].metadata.display == tb1["a"].metadata.display


def test_replace_a_variables_own_value(table_1) -> None:
    tb1 = table_1.copy()
    tb1["a"] = tb1["a"] + 1
    assert (tb1["a"] == pd.Series([2, 3, 4])).all()
    # Metadata for "a" and "b" should be identical.
    assert tb1._fields["a"] == table_1._fields["a"]
    assert tb1._fields["b"] == table_1._fields["b"]

    # Idem using +=.
    tb1 = table_1.copy()
    tb1["a"] += 1
    assert (tb1["a"] == pd.Series([2, 3, 4])).all()
    # Metadata for "a" and "b" should be identical.
    assert tb1._fields["a"] == table_1._fields["a"]
    assert tb1._fields["b"] == table_1._fields["b"]


def test_operations_of_variable_and_scalar(table_1):
    table_1_original = table_1.copy()
    table_1["a"] = table_1["a"] + 1
    table_1["a"] += 1
    table_1["a"] = table_1["a"] - 1
    table_1["a"] -= 1
    table_1["a"] = table_1["a"] * 1
    table_1["a"] *= 1
    table_1["a"] = table_1["a"] / 1
    table_1["a"] /= 1
    table_1["a"] = table_1["a"] // 1
    table_1["a"] //= 1
    table_1["a"] = table_1["a"] % 1
    table_1["a"] %= 1
    table_1["a"] = table_1["a"] ** 1
    table_1["a"] **= 1

    # None of the operations above should have affected the metadata.
    assert table_1["a"].metadata == table_1_original["a"].metadata


def test_create_new_variable_as_product_of_other_two(table_1, sources, origins, licenses) -> None:
    tb1 = table_1.copy()
    tb1["e"] = tb1["a"] * tb1["b"]
    _assert_untouched_data_and_metadata_did_not_change(tb1=tb1, tb1_expected=table_1)
    assert (tb1["e"] == pd.Series([4, 10, 18])).all()
    assert tb1["e"].metadata.title is None
    assert tb1["e"].metadata.description is None
    assert tb1["e"].metadata.description_key == [
        "Key description point 1 of Variable 1",
        "Common key description point",
        "Key description point 1 of Variable 2",
        "Key description point 2 of Variable 2",
    ]
    assert tb1["e"].metadata.sources == [sources[2], sources[1], sources[3]]
    assert tb1["e"].metadata.origins == [origins[2], origins[1], origins[3]]
    assert tb1["e"].metadata.licenses == [licenses[1], licenses[2], licenses[3]]
    # Processing level should be the highest of the two variables.
    assert tb1["e"].metadata.processing_level == "major"
    # Both "a" and "b" have different values in presentation, the combination should have no presentation.
    assert tb1["e"].metadata.presentation is None
    # Since "a" and "b" have identical display, the combination should have the same display.
    assert tb1["e"].metadata.display == tb1["a"].metadata.display


def test_create_new_variable_as_product_of_other_three(table_1, sources, origins, licenses) -> None:
    tb1 = table_1.copy()
    tb1["c"] = tb1["a"] + tb1["b"]
    tb1["f"] = tb1["a"] * tb1["b"] * tb1["c"]
    _assert_untouched_data_and_metadata_did_not_change(tb1=tb1, tb1_expected=table_1)
    assert (tb1["c"] == pd.Series([5, 7, 9])).all()
    assert (tb1["f"] == pd.Series([20, 70, 162])).all()
    assert tb1["f"].metadata.title is None
    assert tb1["f"].metadata.description is None
    assert tb1["f"].metadata.description_key == [
        "Key description point 1 of Variable 1",
        "Common key description point",
        "Key description point 1 of Variable 2",
        "Key description point 2 of Variable 2",
    ]
    assert tb1["f"].metadata.sources == [sources[2], sources[1], sources[3]]
    assert tb1["f"].metadata.origins == [origins[2], origins[1], origins[3]]
    assert tb1["f"].metadata.licenses == [licenses[1], licenses[2], licenses[3]]
    # Processing level should be the highest of all variables.
    assert tb1["c"].metadata.processing_level == "major"
    # Both "a" and "b" have different values in presentation, the combination should have no presentation.
    assert tb1["c"].metadata.presentation is None
    # Since "a" and "b" have identical display (and hence also "c"), the combination should have the same display.
    assert tb1["c"].metadata.display == tb1["a"].metadata.display


def test_create_new_variable_as_division_of_other_two(table_1, sources, origins, licenses) -> None:
    tb1 = table_1.copy()
    tb1["g"] = tb1["a"] / tb1["b"]
    _assert_untouched_data_and_metadata_did_not_change(tb1=tb1, tb1_expected=table_1)
    assert (tb1["g"] == pd.Series([0.25, 0.40, 0.50])).all()
    assert tb1["g"].metadata.title is None
    assert tb1["g"].metadata.description is None
    assert tb1["g"].metadata.description_key == [
        "Key description point 1 of Variable 1",
        "Common key description point",
        "Key description point 1 of Variable 2",
        "Key description point 2 of Variable 2",
    ]
    assert tb1["g"].metadata.sources == [sources[2], sources[1], sources[3]]
    assert tb1["g"].metadata.origins == [origins[2], origins[1], origins[3]]
    assert tb1["g"].metadata.licenses == [licenses[1], licenses[2], licenses[3]]
    # Processing level should be the highest of the two variables.
    assert tb1["g"].metadata.processing_level == "major"
    # Both "a" and "b" have different values in presentation, the combination should have no presentation.
    assert tb1["g"].metadata.presentation is None
    # Since "a" and "b" have identical display, the combination should have the same display.
    assert tb1["g"].metadata.display == tb1["a"].metadata.display


def test_create_new_variable_as_floor_division_of_other_two(table_1, sources, origins, licenses) -> None:
    tb1 = table_1.copy()
    tb1["h"] = tb1["b"] // tb1["a"]
    _assert_untouched_data_and_metadata_did_not_change(tb1=tb1, tb1_expected=table_1)
    assert (tb1["h"] == pd.Series([4, 2, 2])).all()
    assert tb1["h"].metadata.title is None
    assert tb1["h"].metadata.description is None
    # Note that the order of key description points should be first b and then a.
    assert tb1["h"].metadata.description_key == [
        "Key description point 1 of Variable 2",
        "Common key description point",
        "Key description point 2 of Variable 2",
        "Key description point 1 of Variable 1",
    ]
    assert tb1["h"].metadata.sources == [sources[2], sources[3], sources[1]]
    assert tb1["h"].metadata.origins == [origins[2], origins[3], origins[1]]
    assert tb1["h"].metadata.licenses == [licenses[2], licenses[3], licenses[1]]
    # Processing level should be the highest of the two variables.
    assert tb1["h"].metadata.processing_level == "major"
    # Both "a" and "b" have different values in presentation, the combination should have no presentation.
    assert tb1["h"].metadata.presentation is None
    # Since "a" and "b" have identical display, the combination should have the same display.
    assert tb1["h"].metadata.display == tb1["a"].metadata.display


def test_create_new_variable_as_module_division_of_other_two(table_1, sources, origins, licenses) -> None:
    tb1 = table_1.copy()
    tb1["i"] = tb1["a"] % tb1["b"]
    _assert_untouched_data_and_metadata_did_not_change(tb1=tb1, tb1_expected=table_1)
    assert (tb1["i"] == pd.Series([1, 2, 3])).all()
    assert tb1["i"].metadata.title is None
    assert tb1["i"].metadata.description is None
    assert tb1["i"].metadata.description_key == [
        "Key description point 1 of Variable 1",
        "Common key description point",
        "Key description point 1 of Variable 2",
        "Key description point 2 of Variable 2",
    ]
    assert tb1["i"].metadata.sources == [sources[2], sources[1], sources[3]]
    assert tb1["i"].metadata.origins == [origins[2], origins[1], origins[3]]
    assert tb1["i"].metadata.licenses == [licenses[1], licenses[2], licenses[3]]
    # Processing level should be the highest of the two variables.
    assert tb1["i"].metadata.processing_level == "major"
    # Both "a" and "b" have different values in presentation, the combination should have no presentation.
    assert tb1["i"].metadata.presentation is None
    # Since "a" and "b" have identical display, the combination should have the same display.
    assert tb1["i"].metadata.display == tb1["a"].metadata.display


def test_create_new_variable_as_another_variable_to_the_power_of_a_scalar(table_1, sources, origins, licenses) -> None:
    tb1 = table_1.copy()
    tb1["j"] = tb1["a"] ** 2
    _assert_untouched_data_and_metadata_did_not_change(tb1=tb1, tb1_expected=table_1)
    assert (tb1["j"] == pd.Series([1, 4, 9])).all()
    assert tb1["j"].metadata.title == "Title of Table 1 Variable a"
    assert tb1["j"].metadata.description == "Description of Table 1 Variable a"
    assert tb1["j"].metadata.description_key == [
        "Key description point 1 of Variable 1",
        "Common key description point",
    ]
    assert tb1["j"].metadata.sources == [sources[2], sources[1]]
    assert tb1["j"].metadata.origins == [origins[2], origins[1]]
    assert tb1["j"].metadata.licenses == [licenses[1]]
    assert tb1["j"].metadata.processing_level == "minor"
    assert tb1["j"].metadata.presentation == tb1["a"].metadata.presentation
    assert tb1["j"].metadata.display == tb1["a"].metadata.display


def test_create_new_variables_as_another_variable_to_the_power_of_another_variable(
    table_1, sources, origins, licenses
) -> None:
    tb1 = table_1.copy()
    tb1["k"] = tb1["a"] ** tb1["b"]
    _assert_untouched_data_and_metadata_did_not_change(tb1=tb1, tb1_expected=table_1)
    assert (tb1["k"] == pd.Series([1, 32, 729])).all()
    assert tb1["k"].metadata.title is None
    assert tb1["k"].metadata.description is None
    assert tb1["k"].metadata.description_key == [
        "Key description point 1 of Variable 1",
        "Common key description point",
        "Key description point 1 of Variable 2",
        "Key description point 2 of Variable 2",
    ]
    assert tb1["k"].metadata.sources == [sources[2], sources[1], sources[3]]
    assert tb1["k"].metadata.origins == [origins[2], origins[1], origins[3]]
    assert tb1["k"].metadata.licenses == [licenses[1], licenses[2], licenses[3]]
    # Processing level should be the highest of the two variables.
    assert tb1["k"].metadata.processing_level == "major"
    # Both "a" and "b" have different values in presentation, the combination should have no presentation.
    assert tb1["k"].metadata.presentation is None
    # Since "a" and "b" have identical display, the combination should have the same display.
    assert tb1["k"].metadata.display == tb1["a"].metadata.display


def test_get_unique_sources_from_variables(variable_1, variable_2, sources):
    assert get_unique_sources_from_variables([variable_1, variable_2]) == [sources[2], sources[1], sources[3]]
    # Ensure that the function respects the order in which sources appear.
    assert get_unique_sources_from_variables([variable_2, variable_1]) == [sources[2], sources[3], sources[1]]


def test_get_unique_origins_from_variables(variable_1, variable_2, origins):
    assert get_unique_origins_from_variables([variable_1, variable_2]) == [origins[2], origins[1], origins[3]]
    # Ensure that the function respects the order in which origins appear.
    assert get_unique_origins_from_variables([variable_2, variable_1]) == [origins[2], origins[3], origins[1]]


def test_get_unique_licenses_from_variables(variable_1, variable_2, licenses):
    assert get_unique_licenses_from_variables([variable_1, variable_2]) == [licenses[1], licenses[2], licenses[3]]
    # Ensure that the function respects the order in which sources appear.
    assert get_unique_licenses_from_variables([variable_2, variable_1]) == [licenses[2], licenses[3], licenses[1]]


def test_combine_variables_metadata_with_different_fields(variable_1, variable_2, sources, origins, licenses) -> None:
    variable_1 = variable_1.copy()
    variable_2 = variable_2.copy()
    for operation in ["+", "-", "melt", "pivot", "concat"]:
        # TODO: Assert this raises a warning because units are different.
        metadata = combine_variables_metadata([variable_1, variable_2], operation=operation)  # type: ignore
        # If titles/descriptions/units/short_units are different, they should not be propagated.
        assert metadata.title is None
        assert metadata.description is None
        assert metadata.unit is None
        assert metadata.short_unit is None
        assert metadata.sources == [sources[2], sources[1], sources[3]]
        assert metadata.origins == [origins[2], origins[1], origins[3]]
        assert metadata.licenses == [licenses[1], licenses[2], licenses[3]]
        # variable_2 has a major processing level, so the combined variable should have a major processing level.
        assert metadata.processing_level == "major"
        # Both variables have different values in presentation, so the combination should have no presentation.
        assert metadata.presentation is None
        # Since both variables have identical display, the combination should have the same display.
        assert metadata.display == variable_1.metadata.display


def test_combine_variables_metadata_with_equal_fields(variable_1, variable_2) -> None:
    variable_1 = variable_1.copy()
    # Impose that variable 2 is identical to 1.
    variable_2 = variable_1.copy()
    for operation in ["+", "-", "melt", "pivot", "concat"]:
        metadata = combine_variables_metadata([variable_1, variable_2], operation=operation)  # type: ignore
        # If titles/descriptions/units/short_units are identical, they should be propagated.
        assert metadata.title == variable_1.metadata.title
        assert metadata.description == variable_1.metadata.description
        assert metadata.unit == variable_1.metadata.unit
        assert metadata.short_unit == variable_1.metadata.short_unit
        assert metadata.sources == variable_1.metadata.sources
        assert metadata.origins == variable_1.metadata.origins
        assert metadata.licenses == variable_2.metadata.licenses
        # Now both variables have the same processing level, which is minor.
        assert metadata.processing_level == "minor"
        # Both variables have the same presentation, so the combination should have the same presentation.
        assert metadata.presentation == variable_1.metadata.presentation
        # Since both variables have identical display, the combination should have the same display.
        assert metadata.display == variable_1.metadata.display


def test_dropna(table_1) -> None:
    tb1 = table_1.copy()
    tb1.loc[1, "b"] = pd.NA
    new_var = tb1["b"].dropna()
    assert (new_var == pd.Series([4, 6], index=[0, 2])).all()
    # Check that metadata of the new variable coincides with that of the original.
    assert new_var.metadata == tb1["b"].metadata
    # Check that table's metadata is not affected.
    assert tb1.metadata == table_1.metadata
    # Check that the original variable's metadata is not affected.
    assert tb1["b"].metadata == table_1["b"].metadata


def test_copy() -> None:
    v1 = Variable([1, 2, 3], name="dog")
    v1.metadata.title = "dog"
    v1.metadata.license = License(name="dog license")

    v2 = v1.copy()

    # change metadata of a new variable
    v2.metadata.title = "cat"
    v2.metadata.license.name = "cat license"  # type: ignore

    # make sure it doesn't affect original variable
    assert v1.metadata.title == "dog"
    assert v1.metadata.license.name == "dog license"


def test_divide_variables_where_only_numerator_has_metadata(variable_1, variable_2, sources, origins, licenses) -> None:
    variable_1 = variable_1.copy()
    variable_2 = variable_2.copy()
    # We remove metadata values from the denominator.
    variable_2.metadata.title = None
    variable_2.metadata.description = None
    variable_2.metadata.unit = None
    variable_2.metadata.short_unit = None

    # The new variable should have all metadata values from the numerator (for those fields).
    variable = variable_1 / variable_2
    assert variable.metadata.title == variable_1.metadata.title
    assert variable.metadata.description == variable_1.metadata.description
    assert variable.metadata.unit == variable_1.metadata.unit
    assert variable.metadata.short_unit == variable_1.metadata.short_unit
    # Sources, origins and licenses should be propagated as normally.
    assert variable.metadata.sources == [sources[2], sources[1], sources[3]]
    assert variable.metadata.origins == [origins[2], origins[1], origins[3]]
    assert variable.metadata.licenses == [licenses[1], licenses[2], licenses[3]]
    # variable_2 has a major processing level, so the combined variable should have a major processing level.
    assert variable.metadata.processing_level == "major"
    # Both variables have different values in presentation, so the combination should have no presentation.
    assert variable.metadata.presentation is None
    # Since both variables have identical display, the combination should have the same display.
    assert variable.metadata.display == variable_1.metadata.display


def test_divide_variables_where_only_denominator_has_metadata(
    variable_1, variable_2, sources, origins, licenses
) -> None:
    variable_1 = variable_1.copy()
    variable_2 = variable_2.copy()
    # We remove metadata values from the numerator.
    variable_1.metadata.title = None
    variable_1.metadata.description = None
    variable_1.metadata.unit = None
    variable_1.metadata.short_unit = None

    # The new variable should have no metadata values (for those fields).
    variable = variable_1 / variable_2
    assert variable.metadata.title is None
    assert variable.metadata.description is None
    assert variable.metadata.unit is None
    assert variable.metadata.short_unit is None
    # Sources, origins and licenses should be propagated as normally.
    assert variable.metadata.sources == [sources[2], sources[1], sources[3]]
    assert variable.metadata.origins == [origins[2], origins[1], origins[3]]
    assert variable.metadata.licenses == [licenses[1], licenses[2], licenses[3]]
    # variable_2 has a major processing level, so the combined variable should have a major processing level.
    assert variable.metadata.processing_level == "major"
    # Both variables have different values in presentation, so the combination should have no presentation.
    assert variable.metadata.presentation is None
    # Since both variables have identical display, the combination should have the same display.
    assert variable.metadata.display == variable_1.metadata.display


def test_display_propagation_on_divisions(variable_1, variable_2) -> None:
    variable_1 = variable_1.copy()
    variable_2 = variable_2.copy()

    # If the numerator has no display but the denominator has display, the result should have no display.
    variable_1.metadata.display = None
    variable_2.metadata.display = {"numDecimalPlaces": 0}
    variable = variable_1 / variable_2
    assert variable.metadata.display is None

    # If the numerator has display but the denominator has no display, the result should have the numerator's display.
    variable_1.metadata.display = {"numDecimalPlaces": 0}
    variable_2.metadata.display = None
    variable = variable_1 / variable_2
    assert variable.metadata.display == {"numDecimalPlaces": 0}

    # If both numerator and denominator have the same display, the result should have that display.
    variable_1.metadata.display = {"numDecimalPlaces": 0}
    variable_2.metadata.display = {"numDecimalPlaces": 0}
    variable = variable_1 / variable_2
    assert variable.metadata.display == {"numDecimalPlaces": 0}

    # If numerator and denominator have different displays, the result should have no display.
    # NOTE: It is not clear if this is the best choice. Alternatively, we could keep the numerator's display.
    variable_1.metadata.display = {"numDecimalPlaces": 0}
    variable_2.metadata.display = {"numDecimalPlaces": 1}
    variable = variable_1 / variable_2
    assert variable.metadata.display is None


def test_presentation_propagation_on_divisions(variable_1, variable_2) -> None:
    variable_1 = variable_1.copy()
    variable_2 = variable_2.copy()

    # If the numerator has no presentation but the denominator has presentation, the result should have no presentation.
    variable_1.metadata.presentation = None
    variable_2.metadata.presentation = VariablePresentationMeta("test")  # type: ignore
    variable = variable_1 / variable_2
    assert variable.metadata.presentation is None

    # If the numerator has presentation but the denominator has no presentation,
    # the result should have the numerator's presentation.
    variable_1.metadata.presentation = VariablePresentationMeta("test")  # type: ignore
    variable_2.metadata.presentation = None
    variable = variable_1 / variable_2
    assert variable.metadata.presentation == VariablePresentationMeta("test")  # type: ignore

    # If both numerator and denominator have the same presentation, the result should have that presentation.
    variable_1.metadata.presentation = VariablePresentationMeta("test")  # type: ignore
    variable_2.metadata.presentation = VariablePresentationMeta("test")  # type: ignore
    variable = variable_1 / variable_2
    assert variable.metadata.presentation == VariablePresentationMeta("test")  # type: ignore

    # If numerator and denominator have different presentations, the result should have no presentation.
    # NOTE: It is not clear if this is the best choice. Alternatively, we could keep the numerator's presentation.
    variable_1.metadata.presentation = VariablePresentationMeta("test 1")  # type: ignore
    variable_2.metadata.presentation = VariablePresentationMeta("test 2")  # type: ignore
    variable = variable_1 / variable_2
    assert variable.metadata.presentation is None


def test_combine_variables_sort(variable_1, variable_2) -> None:
    variable_1 = variable_1.copy().astype(str) + "a"
    variable_2 = variable_2.copy().astype(str) + "b"

    variable_1.m.type = "ordinal"
    variable_1.m.sort = ["1a", "2a", "3a"]

    assert variable_1.m.type == "ordinal"
    assert variable_1.dropna().m.sort == ["1a", "2a", "3a"]

    variable_2.m.type = "ordinal"
    variable_2.m.sort = ["1b", "2b", "3b"]

    variable_3 = variable_1 + variable_2
    assert variable_3.m.type == "ordinal"
    assert variable_3.m.sort == []


def test_combine_variables_metadata_uses_all_fields() -> None:
    """Make sure we don't forget to process new fields in combine_variables_metadata in case we're adding
    new fields to the VariableMeta.
    If you add a new field to VariableMeta, you should add it to combine_variables_metadata as well and
    this unit test.
    """
    assert {f.name for f in dataclasses.fields(VariableMeta)} == {
        "title",
        "description",
        "description_short",
        "description_from_producer",
        "description_key",
        "sources",
        "origins",
        "licenses",
        "unit",
        "short_unit",
        "display",
        "processing_level",
        "processing_log",
        "presentation",
        "license",
        "type",
        "sort",
        # This is only used right before upserting to grapher
        "additional_info",
        # TODO: we should implement this, if not identical, append one after another
        "description_processing",
        # dimensions
        "dimensions",
        "original_title",
        "original_short_name",
    }


def test_set_categories(variable_3) -> None:
    new_var = variable_3.astype("category")
    new_var = new_var.set_categories(["a", "b", "c", "d"])
    assert new_var.m.title == "Title of Variable 3"
    assert tuple(new_var.cat.categories) == ("a", "b", "c", "d")


def test_variable_rolling(variable_1: Variable):
    v = variable_1.copy()

    rolling = v.rolling(window=1).sum()
    assert rolling.m.title == v.m.title

    # make sure we are not modifying the original table
    rolling.m.title = "new"
    assert v.m.title != "new"


def test_truediv_zero_division() -> None:
    v1 = Variable([0, 1, 2], name="v1")
    v2 = Variable([0, 1, 2], name="v2")
    result = v1 / v2
    assert result.isnull()[0]  # Check that 0/0 results in pandas NaN
    assert not result.isnull()[1]  # Check that 1/1 does not result in NaN
    assert not result.isnull()[2]  # Check that 2/2 does not result in NaN


def test_variable_to_frame() -> None:
    """Test that Variable.to_frame() returns a Table with preserved metadata."""
    from owid.catalog.tables import Table

    # Create a variable with metadata
    v = Variable([1, 2, 3], name="gdp")
    v.metadata.title = "GDP"
    v.metadata.unit = "USD"
    v.metadata.description = "Gross Domestic Product"

    # Convert to frame (should return Table, not DataFrame)
    table = v.to_frame()

    # Check that result is a Table
    assert isinstance(table, Table)

    # Check that it has one column
    assert len(table.columns) == 1
    assert table.columns[0] == "gdp"

    # Check that data is preserved
    assert list(table["gdp"]) == [1, 2, 3]

    # Check that metadata is preserved
    assert table["gdp"].metadata.title == "GDP"
    assert table["gdp"].metadata.unit == "USD"
    assert table["gdp"].metadata.description == "Gross Domestic Product"


def test_variable_to_frame_with_custom_name() -> None:
    """Test that Variable.to_frame() respects custom column name.

    Note: When a custom name is provided, pandas creates a new column with that name.
    Metadata is not automatically copied to the renamed column (consistent with pandas behavior).
    """
    from owid.catalog.tables import Table

    # Create a variable with metadata
    v = Variable([10, 20, 30], name="original_name")
    v.metadata.title = "My Variable"
    v.metadata.unit = "kg"

    # Convert to frame with custom name
    table = v.to_frame(name="custom_name")

    # Check that result is a Table
    assert isinstance(table, Table)

    # Check that column has the custom name
    assert len(table.columns) == 1
    assert table.columns[0] == "custom_name"

    # Check that data is preserved
    assert list(table["custom_name"]) == [10, 20, 30]

    # Note: Metadata is not automatically copied when renaming via name parameter
    # This is consistent with pandas Series.to_frame() behavior
    # If you need to preserve metadata when renaming, copy it manually after conversion



from collections import defaultdict

import pandas as pd
import pytest

from owid.catalog.meta import (
    DatasetMeta,
    License,
    Origin,
    Source,
    TableMeta,
    VariableMeta,
    VariablePresentationMeta,
)
from owid.catalog.tables import Table
from owid.catalog.variables import Variable


@pytest.fixture
def sources():
    sources = {
        1: Source(name="Name of Source 1", description="Description of Source 1"),
        2: Source(name="Name of Source 2", description="Description of Source 2"),
        3: Source(name="Name of Source 3", description="Description of Source 3"),
        4: Source(name="Name of Source 4", description="Description of Source 4"),
    }
    return sources


@pytest.fixture
def origins():
    origins = {
        1: Origin(producer="Producer 1", title="Name of Origin 1", description="Description of Origin 1"),
        2: Origin(producer="Producer 2", title="Name of Origin 2", description="Description of Origin 2"),
        3: Origin(producer="Producer 3", title="Name of Origin 3", description="Description of Origin 3"),
        4: Origin(producer="Producer 4", title="Name of Origin 4", description="Description of Origin 4"),
    }
    return origins


@pytest.fixture
def licenses():
    licenses = {
        1: License(name="Name of License 1", url="URL of License 1"),
        2: License(name="Name of License 2", url="URL of License 2"),
        3: License(name="Name of License 3", url="URL of License 3"),
        4: License(name="Name of License 4", url="URL of License 4"),
    }
    return licenses


@pytest.fixture
def variable_1(sources, origins, licenses):
    v1 = Variable(pd.Series([1, 2, 3]), name="Variable 1")
    v1.metadata.title = "Title of Variable 1"
    v1.metadata.description = "Description of Variable 1"
    v1.metadata.unit = "Unit of Variable 1"
    v1.metadata.unit = "Short unit of Variable 1"
    v1.metadata.sources = [sources[2], sources[1]]
    v1.metadata.origins = [origins[2], origins[1]]
    v1.metadata.licenses = [licenses[1]]
    v1.metadata.processing_level = "minor"
    v1.metadata.presentation = VariablePresentationMeta(title_public="Title of Variable 1")
    v1.metadata.display = {"isProjection": True, "numDecimalPlaces": 1}
    return v1


@pytest.fixture
def variable_2(sources, origins, licenses):
    v2 = Variable(pd.Series([4, 5, 6]), name="Variable 2")
    v2.metadata.title = "Title of Variable 2"
    v2.metadata.description = "Description of Variable 2"
    v2.metadata.unit = "Unit of Variable 2"
    v2.metadata.unit = "Short unit of Variable 2"
    v2.metadata.sources = [sources[2], sources[3]]
    v2.metadata.origins = [origins[2], origins[3]]
    v2.metadata.licenses = [licenses[2], licenses[3]]
    v2.metadata.processing_level = "major"
    v2.metadata.presentation = VariablePresentationMeta(title_public="Title of Variable 2")
    v2.metadata.display = {"isProjection": True, "numDecimalPlaces": 1}
    return v2


@pytest.fixture
def variable_3(origins):
    v3 = Variable(pd.Series(["a", "b", "c"]), name="Variable 3")
    v3.metadata.title = "Title of Variable 3"
    v3.metadata.origins = [origins[2]]
    return v3


@pytest.fixture
def table_1(sources, licenses, origins):
    tb1 = Table({"country": ["Spain", "Spain", "France"], "year": [2020, 2021, 2021], "a": [1, 2, 3], "b": [4, 5, 6]})
    tb1.metadata = TableMeta(
        title="Title of Table 1",
        description="Description of Table 1",
        dataset=DatasetMeta(
            sources=[sources[1], sources[2], sources[3]],
            licenses=[licenses[1], licenses[2], licenses[3]],
        ),
    )
    tb1._fields = defaultdict(
        VariableMeta,
        {
            "country": VariableMeta(title="Country Title", description="Description of Table 1 Variable country"),
            "a": VariableMeta(
                title="Title of Table 1 Variable a",
                description="Description of Table 1 Variable a",
                description_short="Short description of Table 1 Variable a",
                description_key=[
                    "Key description point 1 of Variable 1",
                    "Common key description point",
                ],
                description_from_producer="Common description from producer",
                sources=[sources[2], sources[1]],
                origins=[origins[2], origins[1]],
                licenses=[licenses[1]],
                processing_level="minor",
                presentation=VariablePresentationMeta(title_public="Title of Variable a"),
                display={"isProjection": True, "numDecimalPlaces": 1},
            ),
            "b": VariableMeta(
                title="Title of Table 1 Variable b",
                description="Description of Table 1 Variable b",
                description_short="Short description of Table 1 Variable b",
                description_key=[
                    "Key description point 1 of Variable 2",
                    "Common key description point",
                    "Key description point 2 of Variable 2",
                ],
                description_from_producer="Common description from producer",
                sources=[sources[2], sources[3]],
                origins=[origins[2], origins[3]],
                licenses=[licenses[2], licenses[3]],
                processing_level="major",
                presentation=VariablePresentationMeta(title_public="Title of Variable 2"),
                display={"isProjection": True, "numDecimalPlaces": 1},
            ),
        },
    )
    return tb1


@pytest.fixture
def table_2(sources, licenses):
    tb2 = Table(
        {"country": ["Spain", "France", "France"], "year": [2020, 2021, 2022], "a": [10, 20, 30], "c": [40, 50, 60]}
    )
    tb2.metadata = TableMeta(title="Title of Table 2", description="Description of Table 2")
    tb2._fields = defaultdict(
        VariableMeta,
        {
            "country": VariableMeta(
                title="Country Title",
            ),
            "a": VariableMeta(
                title="Title of Table 2 Variable a",
                description="Description of Table 2 Variable a",
                sources=[sources[2]],
                licenses=[licenses[2]],
                processing_level="major",
                presentation=VariablePresentationMeta(title_public="Title of Variable a"),
                display={"isProjection": True, "numDecimalPlaces": 1, "tolerance": 2},
            ),
            "c": VariableMeta(
                title="Title of Table 2 Variable c",
                description="Description of Table 2 Variable c",
                sources=[sources[2], sources[4]],
                licenses=[licenses[4], licenses[2]],
            ),
        },
    )
    return tb2



#
#  test_meta.py
#

from dataclasses import dataclass
from typing import Any

import pytest
import yaml
from dataclasses_json import dataclass_json

from owid.catalog import meta


def test_dict_mixin():
    @meta.pruned_json
    @dataclass_json
    @dataclass
    class Dog:
        name: str | None = None
        age: int | None = None

        def to_dict(self) -> dict[str, Any]: ...  # type: ignore

    assert Dog(name="fred").to_dict() == {"name": "fred"}
    assert Dog(age=10).to_dict() == {"age": 10}


def test_dict_mixin_nested():
    @meta.pruned_json
    @dataclass_json
    @dataclass
    class Cat:
        name: str | None = None
        age: int | None = None

    @meta.pruned_json
    @dataclass_json
    @dataclass
    class Dog:
        name: str | None = None
        age: int | None = None
        cat: Cat | None = None

        def to_dict(self) -> dict[str, Any]: ...  # type: ignore

    assert Dog(name="fred", cat=Cat(name="cred")).to_dict() == {"name": "fred", "cat": {"name": "cred"}}


def test_empty_dataset_metadata():
    d1 = meta.DatasetMeta()
    assert d1.to_dict() == {"is_public": True, "non_redistributable": False}


def test_dataset_version():
    s1 = meta.Source(name="s1", publication_date="2022-01-01")
    s2 = meta.Source(name="s2", publication_date="2022-01-02")

    assert meta.DatasetMeta(version="1").version == "1"
    assert meta.DatasetMeta(sources=[s1]).version == "2022-01-01"
    assert meta.DatasetMeta(sources=[s1, s2]).version is None
    assert meta.DatasetMeta(version="1", sources=[s1]).version == "1"


def test_to_json():
    meta.Source(name="s1", publication_date="2022-01-01").to_json()  # type: ignore


def test_update_from_yaml(tmp_path):
    d = tmp_path / "sub"
    d.mkdir()
    metapath = d / "meta.yml"

    s1 = meta.Source(name="s1")
    s2 = meta.Source(name="s2")

    # save dictionary to yaml using yaml library
    with open(metapath, "w") as f:
        yaml.dump({"dataset": {"sources": [s2.to_dict()]}}, f)

    d1 = meta.DatasetMeta(sources=[s1])
    with pytest.raises(ValueError):
        d1.update_from_yaml(metapath, if_source_exists="fail")

    d1 = meta.DatasetMeta(sources=[s1])
    d1.update_from_yaml(metapath, if_source_exists="replace")
    assert len(d1.sources) == 1

    d1 = meta.DatasetMeta(sources=[s1])
    d1.update_from_yaml(metapath, if_source_exists="append")
    assert len(d1.sources) == 2


def test_load_license_from_dict():
    d = {
        "url": "https://www.unicef.org/legal#terms-of-use",
    }
    license = meta.License.from_dict(d)
    assert license.url == d["url"]


def test_Origin_date_published():
    assert meta.Origin(producer="p", title="a", date_published="2020-01-01").date_published == "2020-01-01"  # type: ignore
    assert meta.Origin(producer="p", title="a", date_published="2020").date_published == "2020"  # type: ignore
    assert meta.Origin(producer="p", title="a", date_published="latest").date_published == "latest"  # type: ignore
    with pytest.raises(ValueError):
        assert meta.Origin(producer="p", title="a", date_published="nope")  # type: ignore


def test_hash():
    origin_a = meta.Origin("a", "b")
    origin_b = meta.Origin("a", "b")
    origin_c = meta.Origin("a", "c")
    assert origin_a == origin_b
    assert origin_a != origin_c
    assert {origin_a, origin_b, origin_c} == {origin_a, origin_c}

    # test hashing of nested dictionary
    var_a = meta.VariableMeta(display={"d": {"a": 1, "b": 2}})
    var_b = meta.VariableMeta(display={"d": {"a": 1, "b": 2}})
    var_c = meta.VariableMeta(display={"d": {"a": 1, "b": 2, "c": 3}})
    assert var_a == var_b
    assert var_a != var_c


def test_from_dict():
    @dataclass
    class X(meta.MetaBase):
        a: int

    @dataclass
    class Y(meta.MetaBase):
        x_list: list[X] | None = None

    # list of objects should be correctly loaded as that object
    y = Y.from_dict({"x_list": [{"a": 1}]})
    assert isinstance(y.x_list[0], X)


def test_render():
    jinja_title = """
    <% if dim_a == "x" %>Title X<% elif dim_a == "y" %>Title Y<% else %>Default Title<% endif %>
    """.strip()

    var_meta = meta.VariableMeta(title=jinja_title)  # type: ignore
    rendered_meta = var_meta.render(dim_dict={"dim_a": "x"})
    assert isinstance(rendered_meta, meta.VariableMeta)
    assert rendered_meta.title == "Title X"


def test_render_description_key():
    jinja_description_key = [
        "<% if dim_a == 'x' %> Desc x <% endif %>  ",
        "<% if dim_a == 'y' %>  Desc y <% endif %>",
        "Desc z",
    ]

    var_meta = meta.VariableMeta(description_key=jinja_description_key)  # type: ignore
    rendered_meta = var_meta.render(dim_dict={"dim_a": "x"})
    assert isinstance(rendered_meta, meta.VariableMeta)
    assert rendered_meta.description_key == ["Desc x", "Desc z"]


def test_render_with_error():
    jinja_title = """
    <% if dim_a == "x" %>Title X<% elif dim_a == "y" %>Title Y<% else %>Default Title<% endif %>
    """.strip()

    var_meta = meta.VariableMeta(title=jinja_title)  # type: ignore
    with pytest.raises(Exception):
        var_meta.render(dim_dict={"dim_b": "x"})


def test_update_variable_metadata():
    """Test the update_variable_metadata function which handles various post-processing tasks."""
    # Create FaqLink for testing pruning of empty fragment_ids
    empty_faq = meta.FaqLink(gdoc_id="123", fragment_id="   ")
    valid_faq = meta.FaqLink(gdoc_id="456", fragment_id="section-1")

    # Create a presentation with grapher_config containing various fields that need conversion
    presentation = meta.VariablePresentationMeta(
        grapher_config={"map": {"colorScale": {"customNumericValues": "1, 5, 10, 50, 100"}}},
        faqs=[empty_faq, valid_faq],
    )

    # Create variable metadata with fields that need processing
    variable = meta.VariableMeta(
        unit="dollars",
        short_unit="$",
        display={
            "numDecimalPlaces": "2",  # Should be converted to int
            "someOtherField": "value",
        },
        description_key=[
            "Important point",
            "",  # Empty entry should be pruned
            "   ",  # Whitespace-only entry should be pruned
            "Another important point",
        ],
        presentation=presentation,
    )

    # Apply the update function
    updated = meta.update_variable_metadata(variable)
    assert updated.display
    assert updated.presentation
    assert updated.presentation.grapher_config

    # Check that unit and short_unit were copied to display
    assert updated.display["unit"] == "dollars"
    assert updated.display["shortUnit"] == "$"

    # Check numDecimalPlaces was converted to int
    assert updated.display["numDecimalPlaces"] == 2
    assert isinstance(updated.display["numDecimalPlaces"], int)

    # Check empty description_key entries were pruned
    assert updated.description_key == ["Important point", "Another important point"]

    # Check numeric conversions in grapher_config
    color_scale = updated.presentation.grapher_config["map"]["colorScale"]

    # Check string was converted to list
    assert color_scale["customNumericValues"] == [1, 5, 10, 50, 100]

    # Check empty FAQs were pruned
    assert len(updated.presentation.faqs) == 1
    assert updated.presentation.faqs[0].gdoc_id == "456"



#
#  owid.catalog.api.catalogs
#
#  Catalog classes for finding and loading data.
#
from __future__ import annotations

import heapq
import json
import os
import re
import tempfile
from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import Any, Literal, cast
from urllib.parse import urlparse

import numpy as np
import numpy.typing as npt
import pandas as pd
import requests
import structlog
from rapidfuzz import fuzz

from owid.catalog import s3_utils
from owid.catalog.api.utils import (
    INDEX_FORMATS,
    OWID_CATALOG_URI,
    OWID_CATALOG_VERSION,
    PREFERRED_FORMAT,
    S3_OWID_URI,
    SUPPORTED_FORMATS,
)
from owid.catalog.datasets import CHANNEL, Dataset, FileFormat
from owid.catalog.tables import Table

log = structlog.get_logger()


def download_private_file_s3(uri: str, tmpdir: str) -> str:
    """Download private files from S3 to temporary directory."""
    parsed = urlparse(uri)
    base, ext = os.path.splitext(parsed.path)
    s3_utils.download(
        S3_OWID_URI + base + ".meta.json",
        tmpdir + "/data.meta.json",
    )
    s3_utils.download(
        S3_OWID_URI + base + ext,
        tmpdir + "/data" + ext,
    )
    return tmpdir + "/data" + ext


def read_frame(uri: str | Path) -> pd.DataFrame:
    """Read catalog index from various formats."""
    if isinstance(uri, Path):
        uri = str(uri)

    if uri.endswith(".feather"):
        return cast(pd.DataFrame, pd.read_feather(uri))
    elif uri.endswith(".parquet"):
        return cast(pd.DataFrame, pd.read_parquet(uri))
    elif uri.endswith(".csv"):
        return pd.read_csv(uri)

    raise ValueError(f"could not detect format of uri: {uri}")


def save_frame(df: pd.DataFrame, path: str | Path) -> None:
    """Save catalog index in specified format."""
    path = str(path)
    if path.endswith(".feather"):
        df.to_feather(path)
    elif path.endswith(".parquet"):
        df.to_parquet(path)
    elif path.endswith(".csv"):
        df.to_csv(path)
    else:
        raise ValueError(f"could not detect what format to write to: {path}")


def _match_score(
    series: pd.Series,
    query: str,
    mode: Literal["exact", "contains", "regex", "fuzzy"],
    case: bool,
    threshold: int = 70,
) -> npt.NDArray[np.float64]:
    """Calculate match scores for a text column based on search mode.

    Args:
        series: Text column to search
        query: Search pattern
        mode: Matching mode - "exact", "contains", "regex", or "fuzzy"
        case: Case-sensitive matching
        threshold: Score cutoff for fuzzy matching (only used when mode="fuzzy")

    Returns:
        Array of match scores: 0-100 for fuzzy, 0 or 100 for other modes.
    """
    if mode == "fuzzy":
        # Fuzzy matching with similarity scoring (0-100)
        if not case:
            query = query.lower()
            series = series.str.lower()
        return np.array([fuzz.WRatio(query, x, score_cutoff=threshold) for x in series], dtype=np.float64)
    elif mode == "regex":
        # Regex pattern matching (0 or 100)
        matches = series.str.contains(query, regex=True, case=case, na=False)
        return np.where(matches, 100.0, 0.0)
    elif mode == "contains":
        # Substring matching (0 or 100)
        matches = series.str.contains(query, regex=False, case=case, na=False)
        return np.where(matches, 100.0, 0.0)
    else:  # mode == "exact"
        # Exact string matching (0 or 100)
        if case:
            matches = series == query
        else:
            matches = series.str.lower() == query.lower()
        return np.where(matches, 100.0, 0.0)


class CatalogMixin:
    """Abstract catalog interface for finding and loading data."""

    channels: Iterable[CHANNEL]
    frame: "CatalogFrame"
    uri: str

    def find(
        self,
        table: str | None = None,
        namespace: str | None = None,
        version: str | None = None,
        dataset: str | None = None,
        channel: CHANNEL | None = None,
        case: bool = False,
        match: Literal["exact", "contains", "regex", "fuzzy"] = "exact",
        fuzzy_threshold: int = 70,
    ) -> "CatalogFrame":
        """Search catalog for tables matching specified criteria.

        Args:
            table: Table name pattern
            namespace: Namespace filter (exact match only)
            version: Version filter (exact match only)
            dataset: Dataset name pattern
            channel: Channel filter (exact match only)
            case: Case-sensitive search (default: False)
            match: How to match table/dataset names (default: "exact"):
                - "exact": Exact string match
                - "contains": Substring match
                - "regex": Regular expression pattern
                - "fuzzy": Typo-tolerant similarity matching
            fuzzy_threshold: Minimum similarity score 0-100 for fuzzy matching.
                Only used when match="fuzzy". (default: 70)

        Returns:
            CatalogFrame with matching tables, sorted by relevance if match="fuzzy".

        Example:
            ```python
            # Exact match (default)
            catalog.find(table="population")

            # Substring search
            catalog.find(table="pop", match="contains")

            # Regex pattern
            catalog.find(table="pop.*density", match="regex")

            # Fuzzy typo-tolerant search
            catalog.find(table="populaton", match="fuzzy", fuzzy_threshold=80)
            ```
        """
        criteria: npt.ArrayLike = np.ones(len(self.frame), dtype=bool)
        scores: npt.NDArray[np.float64] | None = None

        # Table matching with scoring
        if table:
            table_scores = _match_score(self.frame.table, table, match, case, fuzzy_threshold)
            if match == "fuzzy":
                criteria &= table_scores >= fuzzy_threshold
                scores = table_scores
            else:
                criteria &= table_scores > 0

        # Dataset matching with scoring
        if dataset:
            dataset_scores = _match_score(self.frame.dataset, dataset, match, case, fuzzy_threshold)
            if match == "fuzzy":
                criteria &= dataset_scores >= fuzzy_threshold
                # Average scores if both table and dataset specified
                scores = dataset_scores if scores is None else (scores + dataset_scores) / 2
            else:
                criteria &= dataset_scores > 0

        # Exact match filters
        if namespace:
            criteria &= self.frame.namespace == namespace
        if version:
            criteria &= self.frame.version == version
        if channel:
            if channel not in self.channels:
                raise ValueError(
                    f"You need to add `{channel}` to channels in Catalog init "
                    f"(only `{self.channels}` are loaded now)"
                )
            criteria &= self.frame.channel == channel

        matches = self.frame[criteria]
        if "checksum" in matches.columns:
            matches = matches.drop(columns=["checksum"])

        # Sort by relevance if fuzzy matching was used
        if match == "fuzzy" and scores is not None:
            sort_order = np.argsort(-scores[criteria])
            matches = matches.iloc[sort_order]

        return cast(CatalogFrame, matches)

    def find_one(self, *args: str | None, **kwargs: str | None) -> Table:
        """Find and load a single table matching search criteria."""
        return self.find(*args, **kwargs).load()  # type: ignore

    def find_latest(
        self,
        *args: str | None,
        **kwargs: str | None,
    ) -> Table:
        """Find and load the latest version of a table."""
        frame = self.find(*args, **kwargs)  # type: ignore
        if frame.empty:
            raise ValueError("No matching table found")
        else:
            return cast(Table, frame.sort_values("version").iloc[-1].load())

    def __getitem__(self, path: str) -> Table:
        uri = "/".join([self.uri.rstrip("/"), path])
        for _format in SUPPORTED_FORMATS:
            try:
                return Table.read(f"{uri}.{_format}")
            except Exception:
                continue

        raise KeyError(f"no matching table found at: {uri}")


class LocalCatalog(CatalogMixin):
    """Local filesystem catalog."""

    uri: str

    def __init__(self, path: str | Path, channels: Iterable[CHANNEL] = ("garden",)) -> None:
        self.uri = str(path)
        self.channels = channels
        if self._catalog_exists(channels):
            self.frame = CatalogFrame(self._read_channels(channels))
            self.frame._base_uri = self.path.as_posix() + "/"
        else:
            self.reindex()

    @property
    def path(self) -> Path:
        return Path(self.uri)

    def _catalog_exists(self, channels: Iterable[CHANNEL]) -> bool:
        return all([self._catalog_channel_file(channel).exists() for channel in channels])

    def _catalog_channel_file(self, channel: CHANNEL, format: FileFormat = PREFERRED_FORMAT) -> Path:
        return self.path / f"catalog-{channel}.{format}"

    @property
    def _metadata_file(self) -> Path:
        return self.path / "catalog.meta.json"

    def _read_channels(self, channels: Iterable[CHANNEL]) -> pd.DataFrame:
        """Read selected channels from local path."""
        df = pd.concat([read_frame(self._catalog_channel_file(channel)) for channel in channels])
        df.dimensions = df.dimensions.map(lambda s: json.loads(s) if isinstance(s, str) else s)
        return df

    def iter_datasets(self, channel: CHANNEL, include: str | None = None) -> Iterator[Dataset]:
        to_search = [self.path / channel]
        if not to_search[0].exists():
            return

        re_search = re.compile(include or "")

        while to_search:
            dir = heapq.heappop(to_search)
            if (dir / "index.json").exists() and re_search.search(str(dir)):
                yield Dataset(dir)
                continue

            for child in dir.iterdir():
                if child.is_dir():
                    heapq.heappush(to_search, child)

    def reindex(self, include: str | None = None) -> None:
        """Rebuild the catalog index by scanning the directory tree."""
        index = self._scan_for_datasets(include)

        if include:
            index = self._merge_index(self.frame, index)

        index._base_uri = self.path.as_posix() + "/"
        index.version = index.version.astype(str)
        index.dimensions = index.dimensions.map(lambda s: json.loads(s) if isinstance(s, str) else s)

        self._save_index(index)
        self.frame = index

    @staticmethod
    def _merge_index(frame: "CatalogFrame", update: "CatalogFrame") -> "CatalogFrame":
        """Merge two indexes."""
        return CatalogFrame(
            pd.concat(
                [update, frame.loc[~frame.path.isin(update.path)]],
                ignore_index=True,
            )
        )

    def _save_index(self, frame: "CatalogFrame") -> None:
        """Save all channels to disk in separate catalog files."""
        for channel in self.channels:
            channel_frame = frame.loc[frame.channel == channel].reset_index(drop=True)
            for format in INDEX_FORMATS:
                filename = self._catalog_channel_file(channel, cast(FileFormat, format))
                save_frame(channel_frame, filename)

        self._save_metadata({"format_version": OWID_CATALOG_VERSION})

    def _scan_for_datasets(self, include: str | None = None) -> "CatalogFrame":
        """Scan datasets. You can filter by `include` to get better performance."""
        frames = []
        log.info("reindex.start", channels=self.channels, include=include)
        for channel in self.channels:
            channel_frames = []
            for ds in self.iter_datasets(channel, include=include):
                channel_frames.append(ds.index(self.path))
            frames += channel_frames
            log.info(
                "reindex",
                channel=channel,
                datasets=len(channel_frames),
                include=include,
            )

        df = pd.concat(frames, ignore_index=True)

        keys = ["table", "dataset", "version", "namespace", "channel", "is_public"]
        columns = keys + [c for c in df.columns if c not in keys]

        df.sort_values(keys, inplace=True)  # type: ignore
        df = df.loc[:, columns]

        return CatalogFrame(df)

    def _save_metadata(self, contents: dict[str, Any]) -> None:
        with open(self._metadata_file, "w") as ostream:
            json.dump(contents, ostream, indent=2)


class ETLCatalog(CatalogMixin):
    """Remote HTTP catalog for ETL data."""

    uri: str
    timeout: int

    def __init__(
        self, uri: str = OWID_CATALOG_URI, channels: Iterable[CHANNEL] = ("garden",), timeout: int = 30
    ) -> None:
        self.uri = uri
        self.channels = channels
        self.timeout = timeout
        self.metadata = self._read_metadata(self.uri + "catalog.meta.json", timeout=timeout)
        if self.metadata["format_version"] > OWID_CATALOG_VERSION:
            raise PackageUpdateRequired(
                f"library supports api version {OWID_CATALOG_VERSION}, "
                f"but the remote catalog has version {self.metadata['format_version']} "
                "-- please update"
            )

        self.frame = CatalogFrame(self._read_channels(uri, channels))
        self.frame._base_uri = uri

    @property
    def datasets(self) -> pd.DataFrame:
        return self.frame[["namespace", "version", "dataset"]].drop_duplicates()

    @staticmethod
    def _read_metadata(uri: str, timeout: int = 30) -> dict[str, Any]:
        """Read the metadata JSON blob for this repo."""
        resp = requests.get(uri, timeout=timeout)
        resp.raise_for_status()
        return cast(dict[str, Any], resp.json())

    @staticmethod
    def _read_channels(uri: str, channels: Iterable[CHANNEL]) -> pd.DataFrame:
        """Read selected channels from S3."""
        return pd.concat([read_frame(uri + f"catalog-{channel}.{PREFERRED_FORMAT}") for channel in channels])


class CatalogFrame(pd.DataFrame):
    """DataFrame of catalog search results."""

    _base_uri: str | None = None
    _metadata = ["_base_uri"]

    @property
    def _constructor(self) -> type:
        return CatalogFrame

    @property
    def _constructor_sliced(self) -> Any:
        def build(*args: Any, **kwargs: Any) -> Any:
            c = CatalogSeries(*args, **kwargs)
            c._base_uri = self._base_uri
            return c

        return build

    def load(self) -> Table:
        if len(self) == 1:
            return self.iloc[0].load()  # type: ignore
        elif len(self) == 0:
            raise ValueError("no tables found")
        else:
            raise ValueError(f"only one table can be loaded at once (tables found: {', '.join(self.table.tolist())})")

    @staticmethod
    def create_empty() -> "CatalogFrame":
        return CatalogFrame(
            {
                "namespace": [],
                "version": [],
                "table": [],
                "dimensions": [],
                "path": [],
                "format": [],
            }
        )


class CatalogSeries(pd.Series):
    """Single catalog search result row."""

    _metadata = ["_base_uri"]

    @property
    def _constructor(self) -> type:
        return CatalogSeries

    def load(self) -> Table:
        format = None
        if hasattr(self, "format"):
            format = self.format
        elif hasattr(self, "formats") and (self.formats is not None) and len(self.formats) > 0:
            format = PREFERRED_FORMAT if PREFERRED_FORMAT in self.formats else self.formats[0]

        if self.path and format and self._base_uri:
            with tempfile.TemporaryDirectory() as tmpdir:
                uri = self._base_uri + self.path + "." + format

                if not getattr(self, "is_public", True):
                    uri = download_private_file_s3(uri, tmpdir)

                return Table.read(uri)

        raise ValueError("series is not a table spec")


class PackageUpdateRequired(Exception):
    """Raised when catalog format version is newer than library version."""

    pass



import pandas as pd
import pytest

from owid.catalog import Origin, Table, VariableMeta, VariablePresentationMeta
from owid.catalog.utils import hash_any, remove_details_on_demand, underscore


def test_underscore():
    assert (
        underscore(
            "Proportion of young women and men aged 18‑29 years who experienced sexual violence by age 18",
        )
        == "proportion_of_young_women_and_men_aged_18_29_years_who_experienced_sexual_violence_by_age_18"
    )
    assert (
        underscore(
            "`17.11.1 - Developing countries’ and least developed countries’ share of global merchandise exports (%) - TX_EXP_GBMRCH`"
        )
        == "_17_11_1__developing_countries_and_least_developed_countries_share_of_global_merchandise_exports__pct__tx_exp_gbmrch"
    )
    assert underscore("Urban population") == "urban_population"
    assert underscore("Urban population (% of total population)") == "urban_population__pct_of_total_population"
    assert (
        underscore("Women's share of population ages 15+ living with HIV (%)")
        == "womens_share_of_population_ages_15plus_living_with_hiv__pct"
    )
    assert (
        underscore("Water productivity, total (constant 2010 US$ GDP per cubic meter of total freshwater withdrawal)")
        == "water_productivity__total__constant_2010_usd_gdp_per_cubic_meter_of_total_freshwater_withdrawal"
    )
    assert (
        underscore("Agricultural machinery, tractors per 100 sq. km of arable land")
        == "agricultural_machinery__tractors_per_100_sq__km_of_arable_land"
    )
    assert (
        underscore("GDP per capita, PPP (current international $)")
        == "gdp_per_capita__ppp__current_international_dollar"
    )
    assert (
        underscore("Automated teller machines (ATMs) (per 100,000 adults)")
        == "automated_teller_machines__atms__per_100_000_adults"
    )
    assert (
        underscore("Political regimes - OWID based on Boix et al. (2013), V-Dem (v12), and Lührmann et al. (2018)")
        == "political_regimes__owid_based_on_boix_et_al__2013__v_dem__v12__and_luhrmann_et_al__2018"
    )
    assert (
        underscore("Adjusted savings: particulate emission damage (current US$)")
        == "adjusted_savings__particulate_emission_damage__current_usd"
    )
    assert (
        underscore(
            "Benefit incidence of unemployment benefits and ALMP to poorest quintile (% of total U/ALMP benefits)"
        )
        == "benefit_incidence_of_unemployment_benefits_and_almp_to_poorest_quintile__pct_of_total_u_almp_benefits"
    )
    assert (
        underscore("Business extent of disclosure index (0=less disclosure to 10=more disclosure)")
        == "business_extent_of_disclosure_index__0_less_disclosure_to_10_more_disclosure"
    )
    assert underscore("Firms that spend on R&D (% of firms)") == "firms_that_spend_on_r_and_d__pct_of_firms"
    assert (
        underscore(
            "Wages in the manufacturing sector vs. several food prices in the US – U.S. Bureau of Labor Statistics (2013)"
        )
        == "wages_in_the_manufacturing_sector_vs__several_food_prices_in_the_us__u_s__bureau_of_labor_statistics__2013"
    )
    assert (
        underscore('Tax "composition" –\tArroyo Abad  and P. Lindert (2016)')
        == "tax_composition__arroyo_abad__and_p__lindert__2016"
    )
    assert underscore("20th century deaths in US - CDC") == "_20th_century_deaths_in_us__cdc"
    assert (
        underscore("Poverty rate (<50% of median) (LIS Key Figures, 2018)")
        == "poverty_rate__lt_50pct_of_median__lis_key_figures__2018"
    )
    assert underscore("10") == "_10"
    # Test case from UN SDG data with parentheses and special characters in URLs
    # The input has unicode characters that unidecode converts, potentially leaving parentheses
    assert (
        underscore("14.2.1 - EN_SCP_ECSYBA - Implementation (with parens) - Test https://example.com/#hash(test)")
        == "_14_2_1__en_scp_ecsyba__implementation__with_parens__test_https__example_com_hash__test"
    )
    assert (
        underscore(
            "Indicator 1.5.1: Death rate due to exposure to forces of nature (per 100,000 population) *Estimates reported here are based on a 10-year distributed lag for natural disaster mortality. - Past - Scaled"
        )
        == "indicator_1_5_1__death_rate_due_to_exposure_to_forces_of_nature__per_100_000_population__estimates_reported_here_are_based_on_a_10_year_distributed_lag_for_natural_disaster_mortality__past__scaled"
    )
    # Test case with exclamation mark from UN SDG data - double-encoded UTF-8 produces ! after unidecode
    assert (
        underscore(
            "14.2.1 - EN_SCP_ECSYBA - Implementation and adaptive management - POMIUAC MALAGA-BUENAVENTURA : Plan de OrdenaciÃ³n y Manejo Integrado de la  Unidad Ambiental Costera del Complejo de MÃ¡laga - Buenaventura, https://www.cvc.gov.co"
        )
        == "_14_2_1__en_scp_ecsyba__implementation_and_adaptive_management__pomiuac_malaga_buenaventura__plan_de_ordenacia3n_y_manejo_integrado_de_la__unidad_ambiental_costera_del_complejo_de_ma_laga__buenaventura__https__www_cvc_gov_co"
    )
    assert underscore("a|b") == "a_b"
    assert underscore("$/£ exchange rate") == "dollar_ps_exchange_rate"
    assert (
        underscore("‘cost of basic needs’ approach - share of population below poverty line")
        == "cost_of_basic_needs_approach__share_of_population_below_poverty_line"
    )

    # camelCase to snake_case
    assert underscore("camelCase") == "camelcase"
    assert underscore("camelcase", camel_to_snake=True) == "camelcase"
    assert underscore("camelCase", camel_to_snake=True) == "camel_case"
    assert underscore("CamelCase", camel_to_snake=True) == "camel_case"
    assert underscore("CAMELCase", camel_to_snake=True) == "camel_case"
    assert underscore("camelCase1", camel_to_snake=True) == "camel_case1"
    assert underscore("camelCase_1", camel_to_snake=True) == "camel_case_1"


def test_underscore_table():
    df = pd.DataFrame({"A": [1, 2, 3], "b": [1, 2, 3]})
    df.index.names = ["I"]

    t = Table(df)
    t["A"].metadata.description = "column A"
    t["b"].metadata.description = "column B"

    tt = t.underscore()
    assert list(tt.columns) == ["a", "b"]
    assert tt.index.names == ["i"]

    # original column name is moved to title if we underscored it
    assert tt["a"].metadata.description == "column A"
    assert tt["a"].metadata.title == "A"

    # title is None if underscoring didn't change the name
    assert tt["b"].metadata.description == "column B"
    assert tt["b"].metadata.title is None


def test_underscore_table_collision():
    df = pd.DataFrame({"A__x": [1, 2, 3], "B": [1, 2, 3], "A(x)": [1, 2, 3]})
    t = Table(df)
    t["A__x"].metadata.description = "desc1"
    t["B"].metadata.description = "desc2"
    t["A(x)"].metadata.description = "desc3"

    # raise error by default
    with pytest.raises(NameError):
        t.underscore()

    # add suffix
    tt = t.underscore(collision="rename")
    assert list(tt.columns) == ["a__x_1", "b", "a__x_2"]

    # make sure we retain metadata
    assert tt["a__x_1"].metadata.description == "desc1"
    assert tt["b"].metadata.description == "desc2"
    assert tt["a__x_2"].metadata.description == "desc3"


def test_pruned_json():
    meta = VariableMeta(
        origins=[Origin(title="Title", producer="Producer")],
        presentation=VariablePresentationMeta(title_public="Title public"),
    )
    assert meta.to_dict() == {
        "origins": [{"producer": "Producer", "title": "Title"}],
        "presentation": {"title_public": "Title public"},
    }


def test_hash_any():
    assert hash_any("abc") == 15462616177412505458
    assert hash_any(123) == 123
    assert hash_any({"a": [1, 2, 3]}) == 2363935703804446291
    meta = VariableMeta(
        origins=[Origin(title="Title", producer="Producer")],
        presentation=VariablePresentationMeta(title_public="Title public"),
    )
    assert hash_any(meta) == 6982634015113220894


def test_remove_details_on_demand():
    # Test with single DoD reference
    assert remove_details_on_demand("This is a [description](#dod:something).") == "This is a description."

    # Test with multiple DoD references in same text (the original bug case)
    input_text = (
        "Measured in [terawatt-hours](#dod:watt-hours). "
        "Low-carbon sources correspond to renewables and nuclear power, "
        "that produce significantly less [greenhouse-gas emissions](#dod:ghgemissions) than fossil fuels. "
        "Renewables include solar, wind, hydropower, bioenergy, geothermal, wave, and tidal."
    )
    expected = (
        "Measured in terawatt-hours. "
        "Low-carbon sources correspond to renewables and nuclear power, "
        "that produce significantly less greenhouse-gas emissions than fossil fuels. "
        "Renewables include solar, wind, hydropower, bioenergy, geothermal, wave, and tidal."
    )
    assert remove_details_on_demand(input_text) == expected

    # Test with no DoD references - should return unchanged
    assert remove_details_on_demand("This is plain text.") == "This is plain text."

    # Test with multiple DoDs in same sentence
    assert (
        remove_details_on_demand("The [value](#dod:term1) is the [measurement](#dod:term2) of [something](#dod:term3).")
        == "The value is the measurement of something."
    )

    # Test with DoD at the beginning
    assert (
        remove_details_on_demand("[Temperature](#dod:temp) is measured in celsius.")
        == "Temperature is measured in celsius."
    )

    # Test with DoD at the end
    assert remove_details_on_demand("This measures [energy](#dod:energy)") == "This measures energy"

    # Test that regular markdown links are NOT affected
    assert (
        remove_details_on_demand("See [this link](https://example.com) for more info.")
        == "See [this link](https://example.com) for more info."
    )

    # Test with DoD containing special characters in the keyword
    assert (
        remove_details_on_demand("Measured in [CO₂ emissions](#dod:co2-emissions-detail).")
        == "Measured in CO₂ emissions."
    )

    # Test with multiple DoDs and regular markdown links mixed
    mixed_text = (
        "The [energy](#dod:energy) consumption is measured in [terawatt-hours](#dod:watt-hours). "
        "For more information, see [our methodology](https://example.com/methodology)."
    )
    expected_mixed = (
        "The energy consumption is measured in terawatt-hours. "
        "For more information, see [our methodology](https://example.com/methodology)."
    )
    assert remove_details_on_demand(mixed_text) == expected_mixed

    # Test with brackets in the link text
    assert remove_details_on_demand("This is [data (in brackets)](#dod:something).") == "This is data (in brackets)."

    # Test with empty text
    assert remove_details_on_demand("") == ""

    # Test with DoD containing hyphens, underscores, and colons in keyword
    assert remove_details_on_demand("The [indicator](#dod:complex-key_word:123).") == "The indicator."



#
#  owid.catalog.client.indicators
#
#  Indicators API for semantic search of OWID indicators.
#
from __future__ import annotations

from typing import TYPE_CHECKING

import requests

from owid.catalog.api.models import IndicatorResult, ResponseSet
from owid.catalog.core import CatalogPath

if TYPE_CHECKING:
    from owid.catalog.api import Client
    from owid.catalog.variables import Variable


OWID_SEARCH_API = "https://search.owid.io/indicators"


class IndicatorsAPI:
    """API for semantic search of OWID indicators.

    Uses the search.owid.io service to find indicators using
    natural language queries and vector embeddings.

    Example:
        ```python
        from owid.catalog import Client

        client = Client()

        # Search for indicators
        results = client.indicators.search("solar power generation")
        for ind in results:
            print(f"{ind.title} (score: {ind.score:.2f})")

        # Load the table that contains the indicator of interest
        table = results[0].table

        # Get indicator data directly by URI
        variable = client.indicators.get_data("garden/un/2024-07-12/un_wpp/population#population")

        # Or fetch indicator metadata first
        indicator = client.indicators.fetch("garden/un/2024-07-12/un_wpp/population#population")
        variable = indicator.data
        ```
    """

    BASE_URL = OWID_SEARCH_API

    def __init__(self, client: "Client") -> None:
        """Initialize the IndicatorsAPI.

        Args:
            client: The Client instance.
        """
        self._client = client

    def search(
        self,
        query: str,
        *,
        limit: int = 10,
        show_legacy: bool = False,
        timeout: int | None = None,
    ) -> ResponseSet[IndicatorResult]:
        """Search for indicators using natural language.

        Uses semantic search to find indicators that match the
        meaning of your query, not just keyword matching.

        Args:
            query: Natural language search query
                (e.g., "renewable energy capacity", "child mortality rate").
            limit: Maximum number of results to return. Default 10.
            show_legacy: If True, show pre-ETL indicators only. Default False.
            timeout: HTTP request timeout in seconds. Defaults to client timeout.

        Returns:
            SearchResults containing IndicatorResult objects.

        Example:
            ```python
            # Search for indicators
            results = client.indicators.search("CO2 emissions per capita")

            # View results
            for ind in results:
                print(f"{ind.title}")
                print(f"  Score: {ind.score:.3f}")
                print(f"  Path: {ind.path}")

            # Load data from top result
            indicator = results[0].data
            ```
        """
        params = {
            "query": query,
            "limit": limit,
        }

        resp = requests.get(self.BASE_URL, params=params, timeout=timeout or self._client.timeout)
        resp.raise_for_status()
        data = resp.json()

        results = []
        for r in data.get("results", []):
            path = r.get("catalog_path", "")

            # If legacy indicator, keep only if asked to
            if "/NULL/" in path:
                if not show_legacy:
                    # Skip legacy indicators unless requested
                    continue
                path = None
            results.append(
                IndicatorResult(
                    indicator_id=r.get("indicator_id", 0),
                    title=r.get("title", ""),
                    score=r.get("score", 0.0),
                    path=path,
                    description=r.get("description", ""),
                    column_name=r.get("metadata", {}).get("column", ""),
                    unit=r.get("metadata", {}).get("unit", ""),
                    n_charts=r.get("n_charts", 0),
                )
            )

        return ResponseSet(
            results=results,
            query=query,
            limit=data.get("total_results", len(results)),
        )

    def fetch(self, path: str, *, load_data: bool = False, timeout: int | None = None) -> IndicatorResult:
        """Fetch a specific indicator by catalog path.

        Args:
            path: Catalog path in format "channel/namespace/version/dataset/table#column"
                  (e.g., "grapher/un/2024/pop/population#population_total")
            load_data: If True, preload indicator data immediately.
                       If False (default), data is loaded lazily when accessed via .data property.
            timeout: HTTP request timeout in seconds. Defaults to client timeout.

        Returns:
            IndicatorResult with full details. Access .data to get the variable.

        Raises:
            ValueError: If path format is invalid, table not found, or column doesn't exist.

        Example:
            ```python
            # Fetch indicator by URI
            indicator = client.indicators.fetch("garden/un/2024-07-12/un_wpp/population#population")
            print(f"Title: {indicator.title}")
            variable = indicator.data
            ```

        Note:
            Known limitations of this method:
            - Uses TablesAPI internally, loading the full table to access a single column
            - No indicator_id is assigned when fetching by URI (returns None)
            - Future optimization may allow direct indicator access without full table loading
        """
        effective_timeout = timeout or self._client.timeout

        # Parse path using CatalogPath
        catalog_path = CatalogPath.from_str(path)

        if not catalog_path.variable:
            raise ValueError(
                f"Invalid indicator path format: '{path}'. "
                "Expected format: 'channel/namespace/version/dataset/table#column' (missing #column)"
            )

        # Build table path (without variable)
        table_path = catalog_path.table_path
        column_name = catalog_path.variable

        if table_path is None:
            raise ValueError(
                f"Invalid indicator path format: '{path}'. "
                "Expected format: 'channel/namespace/version/dataset/table#column'"
            )

        # Fetch table header (structure only, no rows) to validate column and extract metadata
        try:
            table_result = self._client.tables.fetch(table_path, timeout=effective_timeout)
            # Load only the header (columns and metadata, no data rows)
            table = table_result.data_header
        except Exception as e:
            raise ValueError(f"Failed to load table at path: '{table_path}'. Error: {e}") from e

        # Verify column exists in table
        if column_name not in table.columns:
            available_cols = ", ".join(f"'{col}'" for col in sorted(table.columns)[:10])
            total_cols = len(table.columns)
            cols_display = f"{available_cols}{'...' if total_cols > 10 else ''}"

            raise ValueError(
                f"Column '{column_name}' not found in table '{table_path}'. "
                f"Available columns ({total_cols}): {cols_display}"
            )

        # Extract metadata from the Variable object
        metadata = table[column_name].metadata

        # Create IndicatorResult
        indicator = IndicatorResult(
            indicator_id=None,  # TODO: No ID when fetching by URI. We should provide one.
            title=metadata.title or column_name,
            score=1.0,  # Direct fetch - perfect match
            path=path,
            description=metadata.description or "",
            column_name=column_name,
            unit=metadata.unit or "",
            n_charts=None,  # TODO: Not available when fetched by URI.
        )

        # Don't store the table - let it be loaded lazily when indicator.data is accessed
        # The table will be loaded fresh via indicator._load_table() when needed

        # Preload data if requested
        if load_data:
            _ = indicator.data

        return indicator

    def get_data(self, path: str, *, timeout: int | None = None) -> "Variable":
        """Fetch indicator data as a Variable.

        Convenience method equivalent to fetch(path).data

        Args:
            path: Catalog path in format "channel/namespace/version/dataset/table#column"
                  (e.g., "grapher/un/2024/pop/population#population_total")
            timeout: HTTP request timeout in seconds. Defaults to client timeout.

        Returns:
            Variable (pandas Series with metadata).

        Raises:
            ValueError: If path format is invalid, table not found, or column doesn't exist.

        Example:
            ```python
            variable = client.indicators.get_data("garden/un/2024-07-12/un_wpp/population#population")
            print(variable.head())
            print(variable.metadata.unit)
            ```
        """
        return self.fetch(path, timeout=timeout).data



#
#  catalogs.py
#  owid-catalog
#
#  DEPRECATED: Use owid.catalog.Client API instead
#
"""Legacy catalog interface (deprecated).

**DEPRECATED**: This module is deprecated. Use the Client API instead.

Migration example:
    ```python
    # OLD (deprecated):
    from owid.catalog import find
    results = find(table="population", namespace="un")
    table = results.iloc[0].load()

    # NEW (recommended):
    from owid.catalog import Client
    client = Client()
    results = client.tables.search(table="population", namespace="un")
    table = results[0].data
    ```

"""

from __future__ import annotations

import warnings
from collections.abc import Iterable
from typing import TYPE_CHECKING

from deprecated import deprecated

# Re-export catalog classes for backwards compatibility
from owid.catalog.api.catalogs import (
    CatalogFrame,
    CatalogSeries,
    ETLCatalog,
    LocalCatalog,
    PackageUpdateRequired,
    read_frame,
    save_frame,
)

# Re-export constants from utils
from owid.catalog.api.utils import INDEX_FORMATS, OWID_CATALOG_URI, S3_OWID_URI

# Re-export constants
from owid.catalog.datasets import CHANNEL, PREFERRED_FORMAT, SUPPORTED_FORMATS, FileFormat

if TYPE_CHECKING:
    from owid.catalog.api import Client

# Global cache for backwards compatibility
REMOTE_CATALOG: ETLCatalog | None = None
_CLIENT_INSTANCE: "Client | None" = None


def _warn_deprecated(func_name: str, alternative: str) -> None:
    """Issue deprecation warning for legacy functions."""
    warnings.warn(
        f"owid.catalog.{func_name}() is deprecated and will be removed in v1.0. "
        f"Use {alternative} instead. "
        f"See: https://docs.owid.io/catalog-api-migration",
        DeprecationWarning,
        stacklevel=3,
    )


def _get_client() -> "Client":
    """Get or create the global Client instance."""
    global _CLIENT_INSTANCE
    if _CLIENT_INSTANCE is None:
        from owid.catalog.api import Client

        _CLIENT_INSTANCE = Client()
    return _CLIENT_INSTANCE


@deprecated(
    version="0.4.0",
    reason="Use Client().tables.search() instead. See: https://docs.owid.io/catalog-api-migration",
)
def find(
    table: str | None = None,
    namespace: str | None = None,
    version: str | None = None,
    dataset: str | None = None,
    channels: Iterable[CHANNEL] = ("garden",),
    case: bool = False,
    regex: bool = True,
    fuzzy: bool = False,
    threshold: int = 70,
) -> CatalogFrame:
    """Search remote catalog (DEPRECATED).

    **DEPRECATED**: Use Client().tables.search() instead.

    Convenience function that searches the default Our World in Data remote
    catalog. Automatically initializes and caches the catalog connection.

    Args:
        table: Table name pattern to search for
        namespace: Filter by namespace (exact match)
        version: Filter by version (exact match)
        dataset: Dataset name pattern to search for
        channels: Data channels to search (default: garden only)
        case: Case-sensitive search (default: False)
        regex: Enable regex patterns in table/dataset (default: True)
        fuzzy: Use fuzzy string matching (default: False)
        threshold: Minimum fuzzy match score 0-100 (default: 70)

    Returns:
        CatalogFrame containing matching tables, sorted by relevance if fuzzy=True.

    Example:
        ```python
        # DEPRECATED - use Client API instead
        from owid.catalog import find
        results = find(table="population")
        results = find(table="populaton", fuzzy=True)  # Fuzzy search

        # RECOMMENDED:
        from owid.catalog import Client
        client = Client()
        results = client.tables.search(table="population")
        ```
    """
    _warn_deprecated("find", "Client().tables.search()")

    # Maintain backwards compatibility by populating REMOTE_CATALOG
    global REMOTE_CATALOG
    if REMOTE_CATALOG is None:
        REMOTE_CATALOG = ETLCatalog(channels=channels)
    else:
        # Add new channels if needed
        current_channels = set(REMOTE_CATALOG.channels)
        new_channels = set(channels) - current_channels
        if new_channels:
            REMOTE_CATALOG = ETLCatalog(channels=tuple(current_channels | set(channels)))

    # Convert old parameters to new match parameter for backwards compatibility
    if fuzzy:
        match = "fuzzy"
    elif regex:
        match = "regex"
    else:
        match = "contains"

    # Use Client API internally with new parameters
    client = _get_client()
    results = client.tables.search(
        table=table,
        namespace=namespace,
        version=version,
        dataset=dataset,
        case=case,
        match=match,
        fuzzy_threshold=threshold,
    )

    # Convert ResponseSet to CatalogFrame for backwards compatibility
    return results.to_catalog_frame()


__all__ = [
    # Classes (backwards compatibility)
    "LocalCatalog",
    "ETLCatalog",
    "CatalogFrame",
    "CatalogSeries",
    "PackageUpdateRequired",
    # Functions (deprecated)
    "find",
    # Constants
    "CHANNEL",
    "OWID_CATALOG_URI",
    "S3_OWID_URI",
    "INDEX_FORMATS",
    "PREFERRED_FORMAT",
    "SUPPORTED_FORMATS",
    "FileFormat",
    # Utilities
    "read_frame",
    "save_frame",
]



#
#  owid.catalog.api.tables
#
#  Tables API for querying and loading tables from the OWID catalog.
#
from __future__ import annotations

import json
from collections.abc import Iterable
from typing import TYPE_CHECKING, Literal, cast

from owid.catalog.api.catalogs import ETLCatalog
from owid.catalog.api.models import ResponseSet, TableResult
from owid.catalog.api.utils import OWID_CATALOG_URI, S3_OWID_URI
from owid.catalog.core import CatalogPath
from owid.catalog.core.paths import VALID_CHANNELS
from owid.catalog.datasets import CHANNEL

if TYPE_CHECKING:
    from owid.catalog.api import Client
    from owid.catalog.tables import Table


class TablesAPI:
    """API for querying and loading tables from the OWID catalog.

    Provides methods to search for tables by various criteria and
    load table data directly from the catalog.

    Example:
        ```python
        from owid.catalog import Client

        client = Client()

        # Search for tables
        results = client.tables.search(table="population", namespace="un")

        # Load the first result
        table = results[0].data

        # Get table data directly by path
        table = client.tables.get_data("garden/un/2024-07-12/un_wpp/population")

        # Or fetch metadata first
        table_result = client.tables.fetch("garden/un/2024-07-12/un_wpp/population")
        print(f"Dataset: {table_result.dataset}, Version: {table_result.version}")
        table = table_result.data  # Lazy-load when needed
        ```
    """

    CATALOG_URI = OWID_CATALOG_URI
    S3_URI = S3_OWID_URI

    def __init__(self, client: Client) -> None:
        self._client = client
        self._catalog: ETLCatalog | None = None

    def _get_catalog(self, timeout: int | None = None) -> ETLCatalog:
        """Get or create the remote catalog with all channels loaded.

        Args:
            timeout: HTTP request timeout in seconds. Defaults to client timeout.
                     Only used during initial catalog loading.
        """
        if self._catalog is None:
            # Load all available channels
            self._catalog = ETLCatalog(
                channels=cast(Iterable[CHANNEL], VALID_CHANNELS), timeout=timeout or self._client.timeout
            )

        return self._catalog

    def _build_query(
        self,
        table: str | None = None,
        namespace: str | None = None,
        version: str | None = None,
        dataset: str | None = None,
        channel: str | None = None,
    ) -> str:
        """Build a descriptive query string from search parameters.

        Args:
            table: Table name pattern
            namespace: Namespace filter
            version: Version filter
            dataset: Dataset name pattern
            channel: Channel filter (defaults to 'garden' if not specified)

        Returns:
            Human-readable query string describing the search parameters.
        """
        query_parts = []
        if table:
            query_parts.append(f"table={table!r}")
        if namespace:
            query_parts.append(f"namespace={namespace!r}")
        if version:
            query_parts.append(f"version={version!r}")
        if dataset:
            query_parts.append(f"dataset={dataset!r}")
        if channel:
            query_parts.append(f"channel={channel!r}")
        else:
            # Default to garden channel
            query_parts.append("channel='garden'")

        return " ".join(query_parts) if query_parts else "all tables"

    def search(
        self,
        table: str | None = None,
        namespace: str | None = None,
        version: str | None = None,
        dataset: str | None = None,
        channel: str | None = None,
        case: bool = False,
        match: Literal["exact", "contains", "regex", "fuzzy"] = "exact",
        fuzzy_threshold: int = 70,
        timeout: int | None = None,
    ) -> ResponseSet[TableResult]:
        """Search the catalog for tables matching criteria.

        Args:
            table: Table name pattern to search for
            namespace: Filter by namespace (exact match)
            version: Filter by version (exact match)
            dataset: Dataset name pattern to search for
            channel: Filter by channel (exact match). Defaults to 'garden' if not specified.
            case: Case-sensitive search (default: False)
            match: How to match table/dataset names (default: "exact"):
                - "exact": Exact string match
                - "contains": Substring match
                - "regex": Regular expression pattern
                - "fuzzy": Typo-tolerant similarity matching
            fuzzy_threshold: Minimum similarity score 0-100 for fuzzy matching.
                Only used when match="fuzzy". (default: 70)
            timeout: HTTP request timeout in seconds for catalog loading. Defaults to client timeout.

        Returns:
            ResponseSet containing matching TableResult objects. If match="fuzzy", results are sorted by relevance score.

        Example:
            ```python
            # Exact match (default) - searches garden channel by default
            results = client.tables.search(table="population")

            # Substring match
            results = client.tables.search(table="pop", match="contains")

            # Regex search
            results = client.tables.search(table="population.*density", match="regex")

            # Fuzzy search sorted by relevance
            results = client.tables.search(table="populaton", match="fuzzy")

            # Case-sensitive fuzzy search with custom threshold
            results = client.tables.search(table="GDP", match="fuzzy", case=True, fuzzy_threshold=85)

            # Filter by namespace and version
            results = client.tables.search(
                table="wdi",
                namespace="worldbank_wdi",
                version="2025-09-08",
            )

            # Search in a specific channel
            results = client.tables.search(
                table="wdi",
                namespace="worldbank_wdi",
                version="2025-09-08",
                channel="meadow",
            )

            # Load a specific result
            tb = results[0].data
            ```
        """
        # Default to garden channel if not specified
        if channel is None:
            channel = "garden"

        catalog = self._get_catalog(timeout=timeout)

        # Use catalog.find() with the new match parameter
        matches = catalog.find(
            table=table,
            namespace=namespace,
            version=version,
            dataset=dataset,
            channel=cast(CHANNEL, channel) if channel else None,
            case=case,
            match=match,
            fuzzy_threshold=fuzzy_threshold,
        )

        # Convert to TableResult objects
        results = []
        for _, row in matches.iterrows():
            dimensions = row.get("dimensions", [])
            if isinstance(dimensions, str):
                dimensions = json.loads(dimensions)

            # Handle formats - could be list, numpy array, or None
            formats_raw = row.get("formats", None)
            if formats_raw is None or (hasattr(formats_raw, "__len__") and len(formats_raw) == 0):
                # Fallback to single format field
                format_single = row.get("format", None)
                formats = [format_single] if format_single else []
            else:
                formats = list(formats_raw) if hasattr(formats_raw, "__iter__") else []

            results.append(
                TableResult(
                    table=row["table"],
                    dataset=row["dataset"],
                    version=str(row["version"]),
                    namespace=row["namespace"],
                    channel=row["channel"],
                    path=row["path"],
                    is_public=row.get("is_public", True),
                    dimensions=list(dimensions) if dimensions is not None else [],
                    formats=formats,
                )
            )

        # Build descriptive query from search parameters
        query = self._build_query(
            table=table,
            namespace=namespace,
            version=version,
            dataset=dataset,
            channel=channel,
        )

        return ResponseSet(
            results=results,
            query=query,
            limit=len(results),
        )

    def fetch(self, path: str, *, load_data: bool = False, timeout: int | None = None) -> TableResult:
        """Fetch table metadata by catalog path (without loading data).

        Returns metadata about the table. Access .data property on the result to
        lazy-load the table data.

        Args:
            path: Full catalog path (e.g., "garden/un/2024-07-12/un_wpp/population").
            load_data: If True, preload table data immediately.
                       If False (default), data is loaded lazily when accessed via .data property.
            timeout: HTTP request timeout in seconds for catalog loading. Defaults to client timeout.

        Returns:
            TableResult with metadata. Access .data to get the table.

        Raises:
            ValueError: If table not found.

        Example:
            ```python
            # Get metadata without loading data
            result = client.tables.fetch("garden/un/2024-07-12/un_wpp/population")
            print(f"Dataset: {result.dataset}, Version: {result.version}")

            # Load data when needed
            tb = result.data
            ```
        """
        # Parse path using CatalogPath
        catalog_path = CatalogPath.from_str(path)

        if not catalog_path.table:
            raise ValueError(f"Invalid path format: {path}. Expected format: channel/namespace/version/dataset/table")

        # Search to get full metadata from catalog
        results = self.search(
            table=catalog_path.table,
            namespace=catalog_path.namespace,
            version=catalog_path.version,
            dataset=catalog_path.dataset,
            channel=catalog_path.channel,
            timeout=timeout,
        )

        if len(results) == 0:
            raise ValueError(f"Table not found: {path}")

        # Get first match (should be exact)
        result = results[0]

        # Preload data if requested
        if load_data:
            _ = result.data  # Access property to trigger loading

        return result

    def get_data(self, path: str, *, timeout: int | None = None) -> "Table":
        """Fetch table data directly.

        Convenience method equivalent to fetch(path).data

        Args:
            path: Full catalog path (e.g., "garden/un/2024-07-12/un_wpp/population").
            timeout: HTTP request timeout in seconds for catalog loading. Defaults to client timeout.

        Returns:
            Table (pandas DataFrame with metadata).

        Raises:
            ValueError: If table not found.

        Example:
            ```python
            tb = client.tables.get_data("garden/un/2024-07-12/un_wpp/population")
            print(tb.head())
            print(tb.metadata.title)
            ```
        """
        return self.fetch(path, timeout=timeout).data



"""Tests for CatalogPath class.

Test coverage:

| Class                    | Description                                                              |
|--------------------------|--------------------------------------------------------------------------|
| TestFromStr              | Parsing catalog paths from strings (full paths, dataset-only, URIs)     |
| TestFromStrValidation    | Validation errors during parsing (invalid channel, version, characters) |
| TestProperties           | Property accessors (dataset_path, table_path, step_uri, short_name)     |
| TestWithMethods          | Immutable builder methods (with_version, with_channel, with_table, etc) |
| TestStringRepresentation | String output (__str__, __repr__, roundtrip parsing)                    |
| TestHashingAndEquality   | Hashing and equality (dict keys, sets, string comparison)               |
| TestParentProperty       | Parent path navigation (removing variable/table components)             |
| TestTruedivOperator      | Path joining with / operator (appending table and variable)             |
| TestFspathProtocol       | os.fspath() and pathlib.Path compatibility                              |
| TestAdditionalCoverage   | Edge cases (all channels, direct construction, immutability)            |
"""

import pytest

from owid.catalog.core.paths import CatalogPath


class TestFromStr:
    """Tests for CatalogPath.from_str parsing."""

    def test_parse_full_path(self) -> None:
        """Parse a complete catalog path with all components."""
        p = CatalogPath.from_str("grapher/who/2024-01-15/gho/life_expectancy#value")
        assert p.channel == "grapher"
        assert p.namespace == "who"
        assert p.version == "2024-01-15"
        assert p.dataset == "gho"
        assert p.table == "life_expectancy"
        assert p.variable == "value"

    def test_parse_without_variable(self) -> None:
        """Parse path without variable component."""
        p = CatalogPath.from_str("grapher/who/2024/gho/life_expectancy")
        assert p.table == "life_expectancy"
        assert p.variable is None

    def test_parse_dataset_only(self) -> None:
        """Parse path with only dataset (no table/variable)."""
        p = CatalogPath.from_str("meadow/un/2024-06-01/population")
        assert p.channel == "meadow"
        assert p.namespace == "un"
        assert p.version == "2024-06-01"
        assert p.dataset == "population"
        assert p.table is None
        assert p.variable is None

    def test_parse_year_only_version(self) -> None:
        """Parse path with year-only version."""
        p = CatalogPath.from_str("garden/worldbank/2024/wdi/gdp#value")
        assert p.version == "2024"

    def test_parse_latest_version(self) -> None:
        """Parse path with 'latest' version."""
        p = CatalogPath.from_str("grapher/owid/latest/key_indicators/population#value")
        assert p.version == "latest"

    def test_parse_with_uri_prefix(self) -> None:
        """Parse path with data:// URI prefix."""
        p = CatalogPath.from_str("data://grapher/who/2024/gho/table#var")
        assert p.channel == "grapher"
        assert p.namespace == "who"
        assert p.table == "table"
        assert p.variable == "var"

    def test_parse_snapshot_uri(self) -> None:
        """Parse snapshot:// URI."""
        p = CatalogPath.from_str("snapshot://who/2024-01-01/data")
        assert p.channel == "snapshot"
        assert p.namespace == "who"
        assert p.version == "2024-01-01"
        assert p.dataset == "data"

    def test_parse_with_hyphens_in_names(self) -> None:
        """Names can contain hyphens."""
        p = CatalogPath.from_str("grapher/energy-data/2024/primary-energy/coal-production#value")
        assert p.namespace == "energy-data"
        assert p.dataset == "primary-energy"
        assert p.table == "coal-production"


class TestFromStrValidation:
    """Tests for validation in from_str."""

    def test_invalid_channel(self) -> None:
        """Reject invalid channel."""
        with pytest.raises(ValueError, match="Invalid channel"):
            CatalogPath.from_str("invalid/who/2024/gho/table#var")

    def test_invalid_version_format(self) -> None:
        """Reject invalid version format."""
        with pytest.raises(ValueError, match="Invalid version"):
            CatalogPath.from_str("grapher/who/2024-1-1/gho/table#var")

    def test_invalid_path_too_few_parts(self) -> None:
        """Reject path with too few parts."""
        with pytest.raises(ValueError, match="Invalid catalog path"):
            CatalogPath.from_str("grapher/who/2024")

    def test_invalid_path_too_many_parts(self) -> None:
        """Reject path with too many parts."""
        with pytest.raises(ValueError, match="Invalid catalog path"):
            CatalogPath.from_str("grapher/who/2024/gho/table/extra")

    def test_invalid_name_characters(self) -> None:
        """Reject names with invalid characters."""
        with pytest.raises(ValueError, match="Invalid namespace"):
            CatalogPath.from_str("grapher/who@org/2024/gho/table")


class TestProperties:
    """Tests for CatalogPath properties."""

    def test_short_name(self) -> None:
        """short_name is alias for dataset."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        assert p.short_name == "gho"
        assert p.short_name == p.dataset

    def test_dataset_path(self) -> None:
        """dataset_path returns path to dataset."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        assert p.dataset_path == "grapher/who/2024/gho"

    def test_table_path(self) -> None:
        """table_path returns path to table."""
        p = CatalogPath.from_str("grapher/who/2024/gho/life_expectancy#value")
        assert p.table_path == "grapher/who/2024/gho/life_expectancy"

    def test_table_path_none_when_no_table(self) -> None:
        """table_path is None when no table."""
        p = CatalogPath.from_str("grapher/who/2024/gho")
        assert p.table_path is None

    def test_step_uri(self) -> None:
        """step_uri returns data:// URI."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        assert p.step_uri == "data://grapher/who/2024/gho"

    def test_table_variable(self) -> None:
        """table_variable returns table#variable slug."""
        p = CatalogPath.from_str("grapher/who/2024/gho/life_expectancy#value")
        assert p.table_variable == "life_expectancy#value"

    def test_table_variable_without_variable(self) -> None:
        """table_variable returns just table when no variable."""
        p = CatalogPath.from_str("grapher/who/2024/gho/life_expectancy")
        assert p.table_variable == "life_expectancy"

    def test_table_variable_none_when_no_table(self) -> None:
        """table_variable is None when no table."""
        p = CatalogPath.from_str("grapher/who/2024/gho")
        assert p.table_variable is None


class TestWithMethods:
    """Tests for with_* builder methods."""

    def test_with_version(self) -> None:
        """with_version creates new path with different version."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        p2 = p.with_version("2025-01-01")
        assert p2.version == "2025-01-01"
        assert p.version == "2024"  # Original unchanged

    def test_with_channel(self) -> None:
        """with_channel creates new path with different channel."""
        p = CatalogPath.from_str("meadow/who/2024/gho/table")
        p2 = p.with_channel("garden")
        assert p2.channel == "garden"
        assert p.channel == "meadow"

    def test_with_variable(self) -> None:
        """with_variable creates new path with different variable."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#old_var")
        p2 = p.with_variable("new_var")
        assert p2.variable == "new_var"
        assert p.variable == "old_var"

    def test_with_variable_requires_table(self) -> None:
        """with_variable raises if no table."""
        p = CatalogPath.from_str("grapher/who/2024/gho")
        with pytest.raises(ValueError, match="Cannot set variable without table"):
            p.with_variable("var")

    def test_with_table(self) -> None:
        """with_table creates new path with different table."""
        p = CatalogPath.from_str("grapher/who/2024/gho/old_table#var")
        p2 = p.with_table("new_table")
        assert p2.table == "new_table"
        assert p2.variable == "var"  # Variable preserved

    def test_with_table_none_clears_variable(self) -> None:
        """Setting table to None also clears variable."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        p2 = p.with_table(None)
        assert p2.table is None
        assert p2.variable is None

    def test_without_variable(self) -> None:
        """without_variable removes variable."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        p2 = p.without_variable()
        assert p2.variable is None
        assert p2.table == "table"  # Table preserved

    def test_without_table(self) -> None:
        """without_table removes table and variable."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        p2 = p.without_table()
        assert p2.table is None
        assert p2.variable is None


class TestStringRepresentation:
    """Tests for __str__ and __repr__."""

    def test_str_full_path(self) -> None:
        """__str__ returns full path."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        assert str(p) == "grapher/who/2024/gho/table#var"

    def test_str_without_variable(self) -> None:
        """__str__ without variable."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table")
        assert str(p) == "grapher/who/2024/gho/table"

    def test_str_dataset_only(self) -> None:
        """__str__ dataset only."""
        p = CatalogPath.from_str("grapher/who/2024/gho")
        assert str(p) == "grapher/who/2024/gho"

    def test_repr(self) -> None:
        """__repr__ includes class name."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        assert repr(p) == "CatalogPath('grapher/who/2024/gho/table#var')"

    def test_roundtrip(self) -> None:
        """Parsing str() output gives equivalent path."""
        original = "grapher/who/2024-01-15/gho/life_expectancy#value"
        p = CatalogPath.from_str(original)
        p2 = CatalogPath.from_str(str(p))
        assert p == p2


class TestHashingAndEquality:
    """Tests for hashing and equality (frozen dataclass)."""

    def test_equality(self) -> None:
        """Equal paths are equal."""
        p1 = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        p2 = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        assert p1 == p2

    def test_inequality(self) -> None:
        """Different paths are not equal."""
        p1 = CatalogPath.from_str("grapher/who/2024/gho/table#var1")
        p2 = CatalogPath.from_str("grapher/who/2024/gho/table#var2")
        assert p1 != p2

    def test_hashable(self) -> None:
        """Can be used as dict key."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        d = {p: "value"}
        assert d[p] == "value"

    def test_usable_in_set(self) -> None:
        """Can be added to set."""
        p1 = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        p2 = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        s = {p1, p2}
        assert len(s) == 1

    def test_equality_with_string(self) -> None:
        """Can compare with string directly."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        assert p == "grapher/who/2024/gho/table#var"
        assert p != "grapher/who/2024/gho/table#other"

    def test_equality_string_both_directions(self) -> None:
        """String comparison works in both directions."""
        p = CatalogPath.from_str("grapher/who/2024/gho")
        assert p == "grapher/who/2024/gho"
        # Note: "string" == CatalogPath won't work (string's __eq__ is called)
        # but CatalogPath == "string" works


class TestParentProperty:
    """Tests for parent property."""

    def test_parent_removes_variable(self) -> None:
        """Parent of path with variable removes variable."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        assert p.parent == "grapher/who/2024/gho/table"
        assert p.parent.variable is None
        assert p.parent.table == "table"

    def test_parent_removes_table(self) -> None:
        """Parent of path with table (no variable) removes table."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table")
        assert p.parent == "grapher/who/2024/gho"
        assert p.parent.table is None

    def test_parent_of_dataset_is_self(self) -> None:
        """Parent of dataset-only path is itself (same object)."""
        p = CatalogPath.from_str("grapher/who/2024/gho")
        assert p.parent == p
        assert p.parent is p  # Same object returned

    def test_parent_chain(self) -> None:
        """Can chain parent calls."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        assert p.parent.parent == "grapher/who/2024/gho"


class TestTruedivOperator:
    """Tests for / operator (path joining)."""

    def test_append_table(self) -> None:
        """Can append table with / operator."""
        p = CatalogPath.from_str("grapher/who/2024/gho")
        p2 = p / "life_expectancy"
        assert p2 == "grapher/who/2024/gho/life_expectancy"
        assert p2.table == "life_expectancy"
        assert p2.variable is None

    def test_append_table_and_variable(self) -> None:
        """Can append table#variable with / operator."""
        p = CatalogPath.from_str("grapher/who/2024/gho")
        p2 = p / "life_expectancy#value"
        assert p2 == "grapher/who/2024/gho/life_expectancy#value"
        assert p2.table == "life_expectancy"
        assert p2.variable == "value"

    def test_append_to_table_raises(self) -> None:
        """Cannot append to path that already has table."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table")
        with pytest.raises(ValueError, match="Cannot append"):
            _ = p / "another"


class TestFspathProtocol:
    """Tests for os.fspath() compatibility."""

    def test_fspath_returns_dataset_path(self) -> None:
        """__fspath__ returns dataset_path for use with Path."""
        import os

        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        assert os.fspath(p) == "grapher/who/2024/gho"

    def test_works_with_pathlib(self) -> None:
        """Can use with pathlib.Path / operator."""
        from pathlib import Path

        base = Path("/data")
        p = CatalogPath.from_str("grapher/who/2024/gho/table#var")
        result = base / p
        assert result == Path("/data/grapher/who/2024/gho")


class TestAdditionalCoverage:
    """Additional edge cases for full coverage."""

    def test_with_namespace(self) -> None:
        """with_namespace creates new path with different namespace."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table")
        p2 = p.with_namespace("un")
        assert p2.namespace == "un"
        assert p.namespace == "who"

    def test_with_dataset(self) -> None:
        """with_dataset creates new path with different dataset."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table")
        p2 = p.with_dataset("wdi")
        assert p2.dataset == "wdi"
        assert p.dataset == "gho"

    def test_from_uri_alias(self) -> None:
        """from_uri is an alias for from_str."""
        p1 = CatalogPath.from_str("data://grapher/who/2024/gho")
        p2 = CatalogPath.from_uri("data://grapher/who/2024/gho")
        assert p1 == p2

    def test_snapshot_private_uri(self) -> None:
        """Parse snapshot-private:// URI."""
        p = CatalogPath.from_str("snapshot-private://who/2024-01-01/data")
        assert p.channel == "snapshot"
        assert p.namespace == "who"

    def test_snapshot_with_variable_raises(self) -> None:
        """Snapshot paths cannot have variables."""
        with pytest.raises(ValueError, match="Snapshot paths cannot have indicators"):
            CatalogPath.from_str("snapshot://who/2024-01-01/data#var")

    def test_direct_construction(self) -> None:
        """Can construct CatalogPath directly."""
        p = CatalogPath(
            channel="garden",
            namespace="un",
            version="2024-01-01",
            dataset="population",
            table="estimates",
            variable="total",
        )
        assert str(p) == "garden/un/2024-01-01/population/estimates#total"

    def test_direct_construction_validation(self) -> None:
        """Direct construction validates inputs."""
        with pytest.raises(ValueError, match="Invalid channel"):
            CatalogPath(
                channel="invalid",  # type: ignore
                namespace="un",
                version="2024",
                dataset="data",
            )

    def test_variable_without_table_raises(self) -> None:
        """Cannot have variable without table."""
        with pytest.raises(ValueError, match="Cannot have variable without table"):
            CatalogPath(
                channel="garden",
                namespace="un",
                version="2024",
                dataset="data",
                table=None,
                variable="var",
            )

    def test_all_channels_valid(self) -> None:
        """All defined channels are accepted."""
        channels = [
            "snapshot",
            "garden",
            "meadow",
            "grapher",
            "open_numbers",
            "examples",
            "explorers",
            "external",
            "multidim",
        ]
        for channel in channels:
            p = CatalogPath.from_str(f"{channel}/ns/2024/ds")
            assert p.channel == channel

    def test_immutability(self) -> None:
        """CatalogPath is immutable (frozen dataclass)."""
        p = CatalogPath.from_str("grapher/who/2024/gho/table")
        with pytest.raises(Exception):  # FrozenInstanceError
            p.version = "2025"  # type: ignore



#
#  owid.catalog.api
#
#  Unified client for all OWID data APIs.
#

from __future__ import annotations

from owid.catalog.api.charts import ChartsAPI
from owid.catalog.api.indicators import IndicatorsAPI
from owid.catalog.api.models import (
    ChartNotFoundError,
    ChartResult,
    IndicatorResult,
    LicenseError,
    PageSearchResult,
    ResponseSet,
    TableResult,
)
from owid.catalog.api.search import SiteSearchAPI
from owid.catalog.api.tables import TablesAPI


class Client:
    """Unified client for all OWID data APIs.

    Provides access to our main APIs:

    - ChartsAPI: Fetch and search for published charts
    - IndicatorsAPI: Semantic search for data indicators
    - TablesAPI: Query and load tables from the data catalog

    Attributes:
        charts: ChartsAPI instance for chart operations and search.
        indicators: IndicatorsAPI instance for indicator search.
        tables: TablesAPI instance for catalog operations.
        datasets: Deprecated alias for tables (backwards compatibility).

    Example:
        ```python
        from owid.catalog import Client

        client = Client()

        # Charts: Published visualizations
        results = client.charts.search("climate change")
        chart = client.charts.fetch("life-expectancy")
        df = client.charts.get_data("animals-slaughtered-for-meat")

        # Tables: Catalog datasets
        results = client.tables.search(table="population", namespace="un")
        table = client.tables.fetch("garden/un/2024-07-12/un_wpp/population")
        tb = client.tables.get_data("garden/un/2024-07-12/un_wpp/population")

        # Indicators: Semantic search for data series
        results = client.indicators.search("renewable energy")
        variable = client.indicators.fetch("garden/un/2024-07-12/un_wpp/population#population")
        data = client.indicators.get_data("garden/un/2024-07-12/un_wpp/population#population")

        ```
    """

    charts: ChartsAPI
    indicators: IndicatorsAPI
    tables: TablesAPI
    datasets: TablesAPI  # Backwards compatibility alias
    timeout: int
    _site_search: SiteSearchAPI

    def __init__(self, timeout: int = 30) -> None:
        """Initialize the client with all API interfaces.

        Args:
            timeout: HTTP request timeout in seconds. Default 30.
        """
        self.timeout = timeout
        self.charts = ChartsAPI(self)
        self.indicators = IndicatorsAPI(self)
        self.tables = TablesAPI(self)
        self.datasets = self.tables  # Backwards compatibility alias
        self._site_search = SiteSearchAPI(self)

    def __repr__(self) -> str:
        return "Client(charts=..., indicators=..., tables=...)"


__all__ = [
    # Main client
    "Client",
    # API classes
    "TablesAPI",
    "ChartsAPI",
    "IndicatorsAPI",
    # Result types for type hints
    "ChartResult",
    "PageSearchResult",
    "IndicatorResult",
    "TableResult",
    "ResponseSet",
    # Exceptions for error handling
    "ChartNotFoundError",
    "LicenseError",
]



#
#  test_datasets.py
#

import json
import os
import random
import shutil
import tempfile
from collections.abc import Iterator
from contextlib import contextmanager
from glob import glob
from os import rmdir
from os.path import exists, join
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pandas as pd
import pytest
import yaml

from owid.catalog import Dataset, DatasetMeta, Table
from owid.catalog.datasets import NonUniqueIndex, PrimaryKeyMissing

from .mocking import mock
from .test_tables import mock_table


def test_dataset_fails_to_load_empty_folder():
    with temp_dataset_dir(create=True) as dirname:
        with pytest.raises(Exception):
            Dataset(dirname)


def test_create_empty():
    with temp_dataset_dir(create=True) as dirname:
        shutil.rmtree(dirname)

        ds = Dataset.create_empty(dirname)

        assert exists(join(dirname, "index.json"))
        with open(join(dirname, "index.json")) as istream:
            doc = json.load(istream)
        assert doc == {"is_public": True, "non_redistributable": False}

        assert len(ds.index()) == 0


def test_create_empty_with_metadata(tmpdir):
    ds = Dataset.create_empty(tmpdir / "dataset", DatasetMeta(namespace="test"))
    assert ds.metadata.namespace == "test"


def test_create_fails_if_non_dataset_dir_exists():
    with temp_dataset_dir(create=True) as dirname:
        with pytest.raises(Exception):
            Dataset.create_empty(dirname)


def test_create_overwrites_entire_folder():
    with temp_dataset_dir(create=True) as dirname:
        with open(join(dirname, "index.json"), "w") as ostream:
            ostream.write('{"clam": "chowder"}')

        with open(join(dirname, "hallo-thar.txt"), "w") as ostream:
            ostream.write("Hello")

    d = Dataset.create_empty(dirname)

    # this should have been deleted
    assert not exists(join(dirname, "hallo-thar.txt"))

    assert (
        open(d._index_file).read().strip()
        == """{
  "is_public": true,
  "non_redistributable": false
}"""
    )


def test_add_table():
    t = mock_table()

    with temp_dataset_dir() as dirname:
        # make a dataset
        ds = Dataset.create_empty(dirname)
        ds.metadata = DatasetMeta(short_name="bob")

        # add the table, it should be on disk now
        ds.add(t)
        assert t.metadata.dataset == ds.metadata

        # check that it's really on disk
        table_files = [
            join(dirname, t.metadata.checked_name + ".feather"),
            join(dirname, t.metadata.checked_name + ".meta.json"),
        ]
        for filename in table_files:
            assert exists(filename)

        # check other methods on Dataset
        assert len(ds) == 1
        assert len(ds.index()) == 1
        assert t.metadata.checked_name in ds

        # load a fresh copy from disk
        t2 = ds[t.metadata.checked_name]
        assert id(t2) != id(t)

        # the fresh copy from disk should be identical to the copy we added
        assert t2.metadata.primary_key == t.metadata.primary_key
        assert t2.equals_table(t)
        assert t2.metadata.dataset == ds.metadata


@patch.dict(os.environ, {})
def test_add_table_without_primary_key():
    t = mock_table().reset_index()

    with temp_dataset_dir() as dirname:
        ds = Dataset.create_empty(dirname)
        with pytest.warns(UserWarning):
            ds.add(t)


@patch.dict(os.environ, {"OWID_STRICT": "1"})
def test_add_table_without_primary_key_strict_mode():
    t = mock_table().reset_index()

    with temp_dataset_dir() as dirname:
        ds = Dataset.create_empty(dirname)
        with pytest.raises(PrimaryKeyMissing):
            ds.add(t)


@patch.dict(os.environ, {"OWID_STRICT": "1"})
def test_add_table_without_unique_index():
    t = Table(pd.DataFrame({"a": [1, 1], "b": [1, 2]}).set_index("a"))

    with temp_dataset_dir() as dirname:
        ds = Dataset.create_empty(dirname)
        with pytest.raises(NonUniqueIndex):
            ds.add(t)


def test_add_table_csv():
    t = mock_table()

    with temp_dataset_dir() as dirname:
        # make a dataset
        ds = Dataset.create_empty(dirname)

        # add the table, it should be on disk now
        ds.add(t, formats=["csv"])

        # check that it's really on disk
        table_files = [
            join(dirname, t.metadata.checked_name + ".csv"),
            join(dirname, t.metadata.checked_name + ".meta.json"),
        ]
        for filename in table_files:
            assert exists(filename)

        # load a fresh copy from disk
        t2 = ds[t.metadata.checked_name]
        assert id(t2) != id(t)

        # the fresh copy from disk should be identical to the copy we added
        assert t2.equals_table(t)


def test_add_table_parquet():
    t = mock_table()

    with temp_dataset_dir() as dirname:
        # make a dataset
        ds = Dataset.create_empty(dirname)

        # add the table, it should be on disk now
        ds.add(t, formats=["parquet"])

        # check that it's really on disk
        assert exists(join(dirname, t.metadata.checked_name + ".parquet"))

        # metadata exists as a sidecar JSON
        assert exists(join(dirname, t.metadata.checked_name + ".meta.json"))

        # load a fresh copy from disk
        t2 = ds[t.metadata.checked_name]
        assert id(t2) != id(t)

        # the fresh copy from disk should be identical to the copy we added
        assert t2.equals_table(t)


def test_metadata_roundtrip():
    with temp_dataset_dir() as dirname:
        d = Dataset.create_empty(dirname)
        d.metadata = mock(DatasetMeta)
        d.save()

        d2 = Dataset(dirname)
        assert d2.metadata == d.metadata


def test_dataset_size():
    with mock_dataset() as d:
        n_expected = len(glob(join(d.path, "*.feather")))
        assert len(d) == n_expected


def test_dataset_iteration():
    with mock_dataset() as d:
        i = 0
        for table in d:
            i += 1
        assert i == len(d)


def test_dataset_hash_changes_with_data_changes():
    with mock_dataset() as d:
        c1 = d.checksum()

        t = mock_table()
        d.add(t)
        c2 = d.checksum()

        assert c1 != c2


def test_dataset_hash_invariant_to_copying():
    # make a mock dataset
    with mock_dataset() as d1:
        # make a copy of it
        with temp_dataset_dir() as dirname:
            d2 = Dataset.create_empty(dirname)
            d2.metadata = d1.metadata
            d2.save()

            for t in d1:
                d2.add(t)

            # the copy should have the same checksum
            assert d2.checksum() == d1.checksum()


def test_snake_case_dataset():
    with mock_dataset() as d:
        # short_name of a dataset must be snake_case
        d.metadata.short_name = "camelCase"
        with pytest.raises(NameError):
            d.save()


def test_snake_case_table():
    with mock_dataset() as d:
        # short_name of a table must be snake_case
        t = mock_table()
        t.metadata.short_name = "camelCase"
        with pytest.raises(NameError):
            d.add(t)

        # short_name of a dataset must be snake_case
        t = mock_table()
        t["camelCase"] = 1
        with pytest.raises(NameError):
            d.add(t)

        # short_name of columns and index names must be snake_case
        t = mock_table()
        t.index.names = ["Country"]
        with pytest.raises(NameError):
            d.add(t)


def test_update_metadata(tmp_path):
    with mock_dataset() as d:
        table_name = d.table_names[0]

        # create test yml file
        temp_file = tmp_path / "my.meta.yml"
        meta = {
            "dataset": {"title": "Dataset title from YAML"},
            "tables": {table_name: {"variables": {"gdp": {"title": "Variable title from YAML"}}}},
        }
        temp_file.write_text(yaml.dump(meta))

        d.update_metadata(temp_file)

        assert d.metadata.title == "Dataset title from YAML"
        assert d[table_name]["gdp"].metadata.title == "Variable title from YAML"


def test_bool():
    with mock_dataset(n_tables=0) as d:
        assert bool(d)


def test_save_fills_channel(tmp_path: Path):
    path = tmp_path / "garden/owid/latest/shortname"
    path.parent.mkdir(exist_ok=True, parents=True)

    d = Dataset.create_empty(path)
    d.metadata = mock(DatasetMeta)
    d.save()

    d2 = Dataset(path)
    assert d2.metadata.channel == "garden"


def test_save_without_valid_channel(tmp_path: Path):
    path = tmp_path / "invalid/owid/latest/shortname"
    path.parent.mkdir(exist_ok=True, parents=True)

    d = Dataset.create_empty(path)
    d.metadata = mock(DatasetMeta)
    d.metadata.channel = None
    d.save()

    d2 = Dataset(path)
    assert d2.metadata.channel is None


@contextmanager
def temp_dataset_dir(create: bool = False) -> Iterator[str]:
    with tempfile.TemporaryDirectory() as dirname:
        if not create:
            rmdir(dirname)
        yield dirname


def create_temp_dataset(dirname: Path | str, n_tables: int | None = None) -> Dataset:
    d = Dataset.create_empty(dirname)
    d.metadata = mock(DatasetMeta)
    d.metadata.version = random.choice(["latest", "2023-01-01"])
    d.metadata.short_name = Path(dirname).name
    d.metadata.is_public = True
    d.save()

    if n_tables is None:
        n_tables = random.randint(2, 5)

    for _ in range(n_tables):
        t = mock_table()
        d.add(t)
    return d


@contextmanager
def mock_dataset(**kwargs: Any) -> Iterator[Dataset]:
    with temp_dataset_dir() as dirname:
        d = create_temp_dataset(dirname, **kwargs)
        yield d



#
#  owid.catalog.client.models
#
#  Pydantic model definitions for API responses.
#
from __future__ import annotations

import io
import json
from typing import TYPE_CHECKING, Any, Callable, Generic, Iterator, TypeVar

import pandas as pd
import requests
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

from owid.catalog.api.utils import _loading_data_from_api
from owid.catalog.core import CatalogPath
from owid.catalog.tables import Table

if TYPE_CHECKING:
    from owid.catalog.api.catalogs import CatalogFrame
    from owid.catalog.variables import Variable

T = TypeVar("T")


class ChartNotFoundError(Exception):
    """Raised when a chart does not exist."""

    pass


class LicenseError(Exception):
    """Raised when chart data cannot be downloaded due to licensing."""

    pass


def _load_table(
    path: str,
    formats: list[str] | None = None,
    is_public: bool = True,
    load_data: bool = True,
) -> "Table":
    """Load a table from the catalog by path.

    Helper function for loading table data. Used by TableResult and IndicatorResult.

    Args:
        path: Table path in catalog (e.g., "grapher/namespace/version/dataset/table")
        formats: List of formats to try. If None, tries all supported formats.
        is_public: Whether the table is publicly accessible.
        load_data: If True, load full data. If False, load only table structure (columns and metadata) without rows.

    Returns:
        Table object with data and metadata (or just metadata if load_data=False).

    Raises:
        KeyError: If no table found at the specified path.
    """
    import tempfile

    from owid.catalog.api.catalogs import download_private_file_s3
    from owid.catalog.api.utils import (
        OWID_CATALOG_URI,
        PREFERRED_FORMAT,
        SUPPORTED_FORMATS,
    )

    # Extract table name for display
    catalog_path = CatalogPath.from_str(path)
    table_name = catalog_path.table or path
    message = f"Loading table '{table_name}'"

    def fct():
        base_uri = OWID_CATALOG_URI
        uri = "/".join([base_uri.rstrip("/"), path])

        # Determine format preference
        if formats:
            formats_to_try = formats
        else:
            formats_to_try = SUPPORTED_FORMATS

        # Prefer feather if available
        if PREFERRED_FORMAT in formats_to_try:
            formats_to_try = [PREFERRED_FORMAT] + [f for f in formats_to_try if f != PREFERRED_FORMAT]

        for fmt in formats_to_try:
            try:
                table_uri = f"{uri}.{fmt}"

                # Handle private files
                if not is_public:
                    tmpdir = tempfile.mkdtemp()
                    table_uri = download_private_file_s3(table_uri, tmpdir)

                # If header_only, return empty table with same structure
                return Table.read(table_uri, load_data=load_data)
            except Exception:
                continue

        raise KeyError(f"No matching table found at: {path}")

    if load_data:
        with _loading_data_from_api(message):
            return fct()
    else:
        return fct()


class ChartResult(BaseModel):
    """An OWID chart (from fetch or search).

    Fields populated depend on the source:
    - fetch(): Provides config and metadata
    - search(): Provides subtitle, available_entities, and num_related_articles

    Core fields (slug, title, url) are always populated.

    Attributes:
        slug: Chart URL identifier (e.g., "life-expectancy").
        title: Chart title.
        url: Full URL to the interactive chart.
        config: Raw grapher configuration dict (from fetch).
        metadata: Chart metadata dict including column info (from fetch).
        subtitle: Chart subtitle/description (from search).
        available_entities: List of entities/countries in the chart (from search).
        num_related_articles: Number of related articles (from search).
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    # Core fields (always present)
    slug: str
    title: str
    url: str

    # From fetch() - full chart details
    config: dict[str, Any] = Field(default_factory=dict)
    metadata: dict[str, Any] = Field(default_factory=dict)

    # From search() - search-specific metadata
    subtitle: str = ""
    available_entities: list[str] = Field(default_factory=list)
    num_related_articles: int = 0

    # Private cached data field
    _data: pd.DataFrame | None = PrivateAttr(default=None)
    _timeout: int = PrivateAttr(default=30)

    @property
    def data(self) -> pd.DataFrame:
        """Lazy-load chart data. Data is cached after first access.

        Returns:
            DataFrame with chart data. Metadata is available in df.attrs.
        """
        if self._data is None:
            self._data = self._load_data()
        return self._data

    def _load_data(self) -> pd.DataFrame:
        """Internal method to fetch chart data."""
        with _loading_data_from_api(f"Fetching chart '{self.slug}'"):
            # Fetch CSV data from the chart
            url = f"https://ourworldindata.org/grapher/{self.slug}.csv?useColumnShortNames=true"
            resp = requests.get(url, timeout=self._timeout)

            if resp.status_code == 404:
                raise ChartNotFoundError(f"No such chart found: {self.slug}")

            if resp.status_code == 403:
                try:
                    error_data = resp.json()
                    raise LicenseError(error_data.get("error", "This chart contains non-redistributable data"))
                except (json.JSONDecodeError, ValueError):
                    raise LicenseError("This chart contains non-redistributable data that cannot be downloaded")

            resp.raise_for_status()

            df = pd.read_csv(io.StringIO(resp.text))

            # Normalize column names
            df = df.rename(columns={"Entity": "entities", "Year": "years", "Day": "years"})
            if "Code" in df.columns:
                df = df.drop(columns=["Code"])

            # Attach metadata
            df.attrs["slug"] = self.slug
            df.attrs["url"] = self.url

            # Rename "years" to "dates" if values are date strings
            if "years" in df.columns and df["years"].astype(str).str.match(r"^\d{4}-\d{2}-\d{2}$").all():
                df = df.rename(columns={"years": "dates"})

            return df


class PageSearchResult(BaseModel):
    """An article/page found via search.

    Attributes:
        slug: Page URL identifier.
        title: Page title.
        url: Full URL to the page.
        excerpt: Short excerpt from the page content.
        authors: List of author names.
        published_at: Publication date string.
        thumbnail_url: URL to thumbnail image.
    """

    slug: str
    title: str
    url: str
    excerpt: str = ""
    authors: list[str] = Field(default_factory=list)
    published_at: str = ""
    thumbnail_url: str = ""


class IndicatorResult(BaseModel):
    """An indicator found via semantic search.

    Attributes:
        title: Indicator title/name.
        indicator_id: Unique indicator ID.
        path: Path in the catalog (e.g., "grapher/un/2024-07-12/un_wpp/population#population").
        channel: Data channel (parsed from path).
        namespace: Data provider namespace (parsed from path).
        version: Version string (parsed from path).
        dataset: Dataset name (parsed from path).
        column_name: Column name in the table.
        description: Full indicator description.
        unit: Unit of measurement.
        score: Semantic similarity score (0-1).
        n_charts: Number of charts using this indicator.
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    # Identification
    title: str
    indicator_id: int | None

    # Location
    path: str | None

    # Structural metadata
    channel: str | None = None
    namespace: str | None = None
    version: str | None = None
    dataset: str | None = None

    # Content metadata
    column_name: str = ""
    description: str = ""
    unit: str = ""

    # Usage metadata
    score: float
    n_charts: int | None = None

    _table: "Table | None" = PrivateAttr(default=None)
    _legacy: bool = PrivateAttr(default=False)

    def model_post_init(self, __context: Any) -> None:
        """Parse dataset, version, namespace, channel from path."""
        if self.path and not self.dataset:
            # Parse using CatalogPath
            try:
                # CatalogPath.from_str() handles the "#" automatically
                parsed = CatalogPath.from_str(self.path)
                # Set parsed fields
                object.__setattr__(self, "dataset", parsed.dataset)
                object.__setattr__(self, "version", parsed.version)
                object.__setattr__(self, "namespace", parsed.namespace)
                object.__setattr__(self, "channel", parsed.channel)
            except Exception:
                # If parsing fails, leave fields empty
                object.__setattr__(self, "_legacy", True)

    @property
    def data(self) -> "Variable":
        """Lazy-load indicator data as a Variable (Series). Data is cached after first access.

        Returns:
            Variable object (pandas Series subclass) with the indicator data.
        """

        if self._table is None:
            self._table = self._load_table()
        # Extract the specific column/variable
        return self._table[self.column_name].dropna()  # type: ignore

    @property
    def table(self) -> "Table":
        """Lazy-load the full table containing this indicator.

        Returns:
            Table object with all columns including this indicator.
        """
        if self._table is None:
            self._table = self._load_table()
        return self._table

    def _load_table(self) -> "Table":
        """Internal method to load the table containing this indicator."""
        # Check path is not None
        if self.path is None:
            raise ValueError("Cannot load table: path is None. Likely because this is a legacy (pre-ETL) table.")
        # Parse path using CatalogPath
        parsed = CatalogPath.from_str(self.path)
        # Use table_path property (without variable)
        if parsed.table_path is None:
            raise ValueError(f"Invalid catalog path: {self.path}")
        return _load_table(parsed.table_path)


class TableResult(BaseModel):
    """A table found in the catalog.

    Attributes:
        table: Table name.
        path: Full path to the table.
        channel: Data channel (garden, meadow, etc.).
        namespace: Data provider namespace.
        version: Version string.
        dataset: Dataset name.
        dimensions: List of dimension columns.
        is_public: Whether the data is publicly accessible.
        formats: List of available formats.
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    # Identification
    table: str

    # Location
    path: str

    # Structural metadata
    channel: str
    namespace: str
    version: str
    dataset: str

    # Content metadata
    dimensions: list[str] = Field(default_factory=list)

    # Technical metadata
    is_public: bool = True
    formats: list[str] = Field(default_factory=list)

    _data: "Table | None" = PrivateAttr(default=None)

    @property
    def data(self) -> "Table":
        """Lazy-load table data. Data is cached after first access.

        Returns:
            Table object with data and metadata.
        """
        if self._data is None:
            self._data = self._load()
        return self._data

    @property
    def data_header(self) -> "Table":
        """Load only the table structure (columns and metadata) without rows.

        This is useful for quickly accessing metadata without loading potentially large datasets.

        Returns:
            Empty Table object with columns and metadata but no data rows.
        """
        return _load_table(self.path, formats=self.formats, is_public=self.is_public, load_data=False)

    def _load(self) -> "Table":
        """Internal method to load table data."""
        return _load_table(self.path, formats=self.formats, is_public=self.is_public)


class ResponseSet(BaseModel, Generic[T]):
    """Generic container for API responses.

    Provides iteration, indexing, and conversion to CatalogFrame
    for backwards compatibility.

    Attributes:
        results: List of result objects.
        query: The query that produced these results.
        limit: Total number of results (may be more than len(results)).
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    results: list[T]
    query: str = ""
    limit: int = 0

    def _get_type_display(self) -> str:
        """Get display name for ResponseSet with generic type."""
        if not self.results:
            return "ResponseSet"

        # Get the type of the first result
        first_result = self.results[0]
        type_name = type(first_result).__name__
        return f"ResponseSet[{type_name}]"

    def model_post_init(self, __context: Any) -> None:
        """Set limit to length of results if not provided."""
        if self.limit == 0:
            self.limit = len(self.results)

    def __iter__(self) -> Iterator[T]:  # type: ignore[override]
        """Iterate over results, not model fields."""
        return iter(self.results)

    def __len__(self) -> int:
        return len(self.results)

    def __getitem__(self, index: int) -> T:
        return self.results[index]

    def __repr__(self) -> str:
        """Display results as a formatted table for better readability."""
        type_display = self._get_type_display()

        if not self.results:
            return f"{type_display}(query={self.query!r}, limit=0, results=[])"

        # Convert to DataFrame for nice tabular display
        df = self.to_frame()

        # Limit display to first 10 rows for readability
        if len(df) == 0:
            return f"{type_display}(query={self.query!r}, limit={self.limit}, results=[])"
        else:
            df_str = str(df)

        # Format as bullet points to show attributes at same level
        # Indent DataFrame lines to align with bullet points
        df_lines = df_str.split("\n")
        indented_df = "\n    ".join(df_lines)

        header = f"{type_display}\n.query={self.query!r}\n.limit={self.limit}\n.results:\n    {indented_df}"

        # Add helper tip at the end
        tip = "\n\nTip: Use .to_frame() for pandas operations, or .latest(by='field') to get most recent"
        return header + tip

    def __str__(self) -> str:
        """Use the same representation for str() and repr()."""
        return self.__repr__()

    def _repr_html_(self) -> str:
        """Display as HTML table in Jupyter notebooks."""
        type_display = self._get_type_display()

        if not self.results:
            return f"<p>{type_display}(query={self.query!r}, limit=0, results=[])</p>"

        df = self.to_frame()
        df_html = df._repr_html_()

        # Format as bullet points to show attributes at same level
        html = f"""<div>
  <p><strong>{type_display}</strong></p>
  <ul style="list-style-type: none; padding-left: 1em;">
    <li><strong>.query</strong>: {self.query!r}</li>
    <li><strong>.limit</strong>: {self.limit}</li>
    <li><strong>.results</strong>:
      <div style="margin-left: 1.5em; margin-top: 0.5em;">
        {df_html}
      </div>
    </li>
  </ul>
</div>"""
        return html

    def to_frame(self) -> pd.DataFrame:
        """Convert results to a DataFrame.

        Returns:
            DataFrame with one row per result.
        """
        if not self.results:
            return pd.DataFrame()

        # Convert Pydantic models to dicts
        rows = []
        for r in self.results:
            if isinstance(r, BaseModel):
                # For ChartResult, exclude large dict fields for better display
                if isinstance(r, ChartResult):
                    row = {
                        "slug": r.slug,
                        "title": r.title,
                        "subtitle": r.subtitle,
                        "url": r.url,
                        "num_related_articles": r.num_related_articles,
                        # Only show count of entities, not full list
                        "num_entities": len(r.available_entities),
                    }
                else:
                    row = r.model_dump()
                rows.append(row)
            else:
                rows.append(r)

        return pd.DataFrame(rows)

    def to_catalog_frame(self) -> "CatalogFrame":
        """Convert to CatalogFrame for backwards compatibility.

        Only works for TableResult and IndicatorResult types.

        Returns:
            CatalogFrame that can use .load() method.
        """
        from owid.catalog.api.catalogs import CatalogFrame as CF
        from owid.catalog.api.utils import OWID_CATALOG_URI

        if not self.results:
            return CF.create_empty()

        # Check result type
        first = self.results[0]
        if isinstance(first, TableResult):
            rows = []
            for r in self.results:
                rows.append(
                    {
                        "table": r.table,  # type: ignore
                        "dataset": r.dataset,  # type: ignore
                        "version": r.version,  # type: ignore
                        "namespace": r.namespace,  # type: ignore
                        "channel": r.channel,  # type: ignore
                        "path": r.path,  # type: ignore
                        "is_public": r.is_public,  # type: ignore
                        "dimensions": r.dimensions,  # type: ignore
                        "format": r.formats[0] if r.formats else "feather",  # type: ignore
                    }
                )
            frame = CF(rows)
            frame._base_uri = OWID_CATALOG_URI
            return frame

        elif isinstance(first, IndicatorResult):
            rows = []
            for r in self.results:
                # Parse catalog path using CatalogPath
                try:
                    parsed = CatalogPath.from_str(r.path)  # type: ignore
                    indicator = parsed.variable or ""
                    channel = parsed.channel
                    namespace = parsed.namespace
                    version = parsed.version
                    dataset = parsed.dataset
                    table = parsed.table or dataset
                    # Use table_path property (without variable)
                    path_part = parsed.table_path or parsed.dataset_path
                except Exception:
                    # Fallback if parsing fails
                    indicator = channel = namespace = version = dataset = table = ""
                    path_part = r.path.split("#")[0] if "#" in r.path else r.path  # type: ignore

                rows.append(
                    {
                        "indicator_title": r.title,  # type: ignore
                        "indicator": indicator,
                        "score": r.score,  # type: ignore
                        "table": table,
                        "dataset": dataset,
                        "version": version,
                        "namespace": namespace,
                        "channel": channel,
                        "is_public": True,
                        "path": path_part,
                        "format": "parquet",
                    }
                )
            frame = CF(rows)
            frame._base_uri = OWID_CATALOG_URI
            return frame

        else:
            raise TypeError(f"Cannot convert {type(first).__name__} results to CatalogFrame")

    def filter(self, predicate: Callable[[T], bool]) -> "ResponseSet[T]":
        """Filter results by predicate function.

        Returns a new ResponseSet with only items that match the predicate.
        The predicate should return True for items to keep.

        Args:
            predicate: Function that takes an item and returns True/False.

        Returns:
            New ResponseSet with filtered results.

        Example:
            ```py
            >>> # Filter results by version
            >>> results.filter(lambda r: r.version > '2024')

            >>> # Filter by namespace
            >>> results.filter(lambda r: r.namespace == "worldbank")

            >>> # Chain multiple filters
            >>> results.filter(lambda r: r.version > '2024').filter(lambda r: r.namespace == "un")
            ```
        """
        filtered_results = [item for item in self.results if predicate(item)]
        return ResponseSet(
            results=filtered_results,
            query=self.query,
            limit=len(filtered_results),
        )

    def sort_by(self, key: str | Callable[[T], Any], *, reverse: bool = False) -> "ResponseSet[T]":
        """Sort results by attribute name or key function.

        Returns a new ResponseSet with items sorted by the specified key.

        Args:
            key: Either an attribute name (string) or a function that extracts a comparison key from each item.
            reverse: If True, sort in descending order (default: False).

        Returns:
            New ResponseSet with sorted results.

        Example:
            ```py
            >>> # Sort by version (ascending)
            >>> results.sort_by('version')

            >>> # Sort by version (descending - latest first)
            >>> results.sort_by('version', reverse=True)

            >>> # Sort by custom function (e.g., by score)
            >>> results.sort_by(lambda r: r.score, reverse=True)

            >>> # Chain sorting and filtering
            >>> results.filter(lambda r: r.version > '2024').sort_by('version', reverse=True)
            ```
        """
        if isinstance(key, str):
            # Sort by attribute name
            sorted_results = sorted(self.results, key=lambda item: getattr(item, key), reverse=reverse)
        else:
            # Sort by key function
            sorted_results = sorted(self.results, key=key, reverse=reverse)

        return ResponseSet(
            results=sorted_results,
            query=self.query,
            limit=self.limit,
        )

    def latest(self, by: str = "version") -> T:
        """Get the most recent result by a specific field.

        Returns the single item with the highest value for the specified field.

        Args:
            by: Attribute name to sort by (default: 'version').
                Common values: 'version', 'published_at', 'score'.

        Returns:
            Single item with the highest value for the specified field.

        Raises:
            ValueError: If no results are available.
            AttributeError: If the specified attribute doesn't exist on the results.

        Example:
            ```py
            >>> # For TableResult - use version (default)
            >>> latest_table = results.latest()
            >>> tb = latest_table.data

            >>> # For IndicatorResult - use version (default)
            >>> latest_indicator = results.latest()
            ```

        Note:
            Support for charts is coming soon.
        """
        if not self.results:
            raise ValueError("No results available to get latest from")

        # Check if attribute exists on first item
        if not hasattr(self.results[0], by):
            # Get available attributes (exclude private ones)
            available = [
                k for k in dir(self.results[0]) if not k.startswith("_") and not callable(getattr(self.results[0], k))
            ]
            raise AttributeError(
                f"Results don't have '{by}' attribute. " f"Available attributes: {', '.join(sorted(available))}"
            )

        return max(self.results, key=lambda item: getattr(item, by))

    def first(self, n: int = 1) -> T | "ResponseSet[T]":
        """Get the first n results.

        Args:
            n: Number of results to return (default: 1).

        Returns:
            If n=1, returns a single item (or None if no results).
            If n>1, returns a new ResponseSet with the first n results.

        Example:
            ```python
            >>> # Get first result
            >>> first_result = results.first()
            >>> tb = first_result.data

            >>> # Get first 5 results
            >>> top_five = results.first(5)

            >>> # Combine with sorting
            >>> latest_five = results.sort_by('version', reverse=True).first(5)
            ```
        """
        if n == 1:
            return self.results[0] if self.results else None  # type: ignore
        else:
            return ResponseSet(
                results=self.results[:n],
                query=self.query,
                limit=self.limit,
            )
